/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 */

#ifndef CHILDSCOPE_H
#define CHILDSCOPE_H

#include <functional>

#include <unity-scopes.h>
#include <unity/scopes/ScopeMetadata.h>
#include <unity/scopes/RegistryProxyFwd.h>

#include <QDebug>
#include <QString>

namespace us = unity::scopes;


/**
    \brief The child scope to be aggregated
    Provides methods to enable common child scope customizations
*/
class AggChildScope
{
public:
    /**
     \brief Constructor
     \param id the fully qualified scope id (for example: com.canonical.scopes.photos-local_photos_local)

    */
    AggChildScope(const std::string id);

    void set_local_id(const std::string local_id);

    std::string local_id();

    /**
     \brief Set our poinnter to the scope's ScopeMetadata
    \param ScopeMetadata
    */
    void set_metadata(us::ScopeMetadata const &);

    /**
    \brief Return a pointer to the scope's ScopeMetadata
    \return Pointer to ScopeMetadata
    */
    std::shared_ptr<us::ScopeMetadata> metadata();

    /**
     \brief Set the displayed category title. Not defined when using_category_title_incoming.
    \param title_ The title
    */
    void set_category_title(const std::string title_);

    /**
    \brief Return the Category title. Not defined when using_category_title_incoming.
    \return title
    */
    std::string category_title();

    /**
     \brief The displayed category title
    \param title_ The title
    */
    void set_category_title_incoming(bool b);

    /**
    \brief Whether category title to be set by first appropriate incoming result.
    \return Return whether category title to be set by first appropriate incoming result.
    */
    bool using_category_title_incoming();

    /**
     \brief Use incoming scope's displayname as category title
    \param true to use displayname (false needed because default value is false)
    */
    void set_category_title_display_name(bool b);

    /**
    \brief Whether category title to be set from scope's display name.
    \return Return whether category title to be set by scope's display name.
    */
    bool using_category_title_display_name();

    /**
    \brief Set to false in first result processing so we can know if the current result is the first. is_first_result is true by default. Only applicable when agg scope creates category for results
    \param Set to false at end of first result processing.
    */
    void set_is_first_result(bool b)
    {
        is_first_result_ = b;
    }

    /**
    \brief Whether the current result is first result only when agg scope creates category for results
    \return Returns whether the current result is the firist result.
    */
    bool is_first_result()
    {
        return is_first_result_;
    }

    /**
    \brief Set to true in first result processing and to false in second result processing so we can know if the current result is the second. is_second_result returns false initially. Only applicable when agg scope creates category for results
    \param Sets wther the result is the second result.
    */
    void set_is_second_result(bool b)
    {
        is_second_result_ = b;
    }

    /**
    \brief Whether the current result is second result only when agg scope creates category for results
    \return Returns whether the current result is the second result.
    */
    bool is_second_result()
    {
        return is_second_result_;
    }

    /**
    \brief Set Category
    \return Category
    */
    void set_category(us::Category::SCPtr category_);

    /**
    \brief Return SCPtr to Category
    \return Category
    */
    us::Category::SCPtr category();

    /**
    \brief The json defining the CategoryRenderer json used when surfacing the first result. You must also declare surface_template to handle the rest.
    */
    void set_first_result_template(const std::string first_result_template);

    /**
    \brief Return the json defining the CategoryRenderer template used when surfacing first result.
    */
    std::string first_result_template();
    /**
 *
    \brief The json defining the CategoryRenderer template used when surfacing results.
    \param surface_template The json passed to the CategoryRenderer
    */
    void set_surface_template(const std::string surface_template);

    /**
    \brief Return the json defining the CategoryRenderer template used when surfacing results.
    */
    std::string surface_template();

    void set_first_common_template_id(std::string id);
    std::string first_common_template_id();

    void set_surface_common_template_id(std::string id);
    std::string surface_common_template_id();

    /**
    \brief Sets whether this scope overrides the category render via json.
    \param Whether to override. Undefined when json does not declare surface_template
    */
    void set_override_surface_template(bool);

    /**
    \brief Whether this scope overrides the category render via json.
    \param Whether to override.
    */
    bool override_surface_template();

    /**
    \brief Return the json defining the CategoryRenderer template used when displaying search results.
    */
    std::string search_template();

    /**
    \brief The json defining the CategoryRenderer template used when displaying search results.
    \param search_template The json passed to the CategoryRenderer
    */
    void set_search_template(const std::string search_template);

    /**
    \brief Sets whether this scope overrides the category render via json in search mode.
    \param Whether to override. Undefined when json does not declare surface_template
    */
    void set_override_search_template(bool);

    /**
    \brief Whether this scope overrides the category render via json when searching.
    \param Whether to override.
    */
    bool override_search_template();

    /**
    \brief Display child scope results only from the specific department
    \param department_id The ID of the child scope department.
    */
    void set_child_department(const std::string department_id_);

    /**
    \brief Return child department id
    \return Return child department id
    */
    std::string child_department();

    /**
    \brief Places this scope in the department whose id is passed.
    \param department_id The ID of the department.
    */
    void set_department(const std::string department_id_);

    /**
    \brief Return scope's department id
    \return Return scope's department id
    */
    std::string department();

    /**
    \brief Display child scope results only from the specified category
    \parm category_id The ID of the cagtegory of the result.
    */
    void set_child_category(const std::string category_id);

    /**
    \brief Return the category_id. Undefined it not using set_child_category()
    \return The child_category
    */
    std::string child_category();
    /**
    \brief Limit child scope results to this number.
    Note: Only use this when also using set_child_category(), otherwise use set_cardinality()
    \parm max The maximum number of results to display.
    */
    void set_child_category_max_results(int max);

    /**
    \brief Return whether scope is using child_category_max_results.
    \return Whether scope is using child_category_max_results.
    */
    bool using_child_category_max_results();

    /**
    \brief Return the child_category_max_results.
    \return Return child_category_max_results.
    */
    int child_category_max_results();

    /**
    \brief Limit child scope results to this number through scope cardinality.
    Note: Only use this when not using set_child_category(), otherwise use set_category_max_results()
    \parm cardinality The maximum number of results to display.
    */
    void set_cardinality(int cardinality);

    /**
    \brief Return scope cardinality
    \return cardinality
    */
    int cardinality();

    /**
    \brief Return scope ID
    \return The scope ID.
    */
    std::string id();

    /**
    \brief Determines whether category has  link to the child scope
    \param True to have link, false to not have link.
    */
    void set_category_link_to_child(bool b);

    /**
    \brief Return True if scope is using categoryLinkToChild, else false 
    \return The scope ID.
    */
    bool category_link_to_child();

    /**
    \brief Return the CannedQuery created when category_link_to_child is set. Undefined if category_link-to-child is not set.
    \return The CannedQuery.
    */
    std::shared_ptr<us::CannedQuery> canned_query();

    /**
    \brief Return True if the AggChildScope scope is installed and registered by id().
    \return True when scope is present.
    */
    bool exists()
    {
        return exists_;
    }

    /**
    \brief Find out whether the scope is enabled in settings
    \return True when scope is enabled.
    */
    bool enabled(const us::VariantMap settings);

    /**
    \brief Return this scope by ScopeProxy
    */
    us::ScopeProxy proxy();

    /**
    \brief This is called automatically. You generally do not need to call this. If the scope is in the registry, its proxy is saved and its
    \param A RegistryProxy
    */
    void set_proxy(us::RegistryProxy const);

    /**
    \brief This is called automatically and you generally do not need to call this. Increments the results_idx;
    */
    void inc_result_idx();

    /**
    \brief This is called automatically and you generally do not need to call this. ;
    \return The result_idx`
    */
    int result_idx();

    void set_buffer(bool);
    bool buffer();

    void set_category_scope(bool);

    bool category_scope();

    void set_category_id(std::string);

    std::string category_id();

    void set_keyword_scope(bool);

    bool keyword_scope();

    void set_keyword(std::string);

    std::string keyword();

    void set_keyword_scope_shared_cat_name(std::string);

    std::string keyword_scope_shared_cat_name();

    bool keyword_uses_shared_cat();

    bool source_finder();
    void set_source_finder(bool);

    bool only_in_search();
    void set_only_in_search(bool);

private:

    std::shared_ptr<us::ScopeMetadata> scope_metadata_;

    bool buffer_ = true;
    bool category_scope_ = false;
    std::string category_scope_id_;
    bool keyword_scope_ = false;
    std::string keyword_;
    std::string keyword_scope_shared_cat_name_;
    bool uses_shared_cat_ = false;
    std::string id_;
    std::string local_id_;
    std::string cat_title = "";
    bool using_category_link_to_child;
    bool using_category_title_incoming_ = false;
    bool using_category_title_display_name_ = false;
    bool is_first_result_ = true;
    bool is_second_result_ = false;
    std::string dept_id = "";
    std::string subsearch_dept_id = "";
    bool using_category = false;
    bool only_in_search_ = false;
    bool using_child_category_max_results_ = false;
    int child_category_max_results_ = 6;
    int result_idx_ = 0;
    std::string category_id_ = "";
    std::string query_string;
    int cardinality_ = 5;//a default quantity required
    us::ScopeProxy proxy_;
    bool exists_ = false;
    std::shared_ptr<us::CannedQuery> canned_query_;
    us::Category::SCPtr cat;
    bool override_surface_template_ = true;
    bool override_search_template_ = true;
    std::string TEMPLATE_NOT_LOGGED_IN;
    std::string TEMPLATE_FIRST_RESULT;
    std::string first_common_template_id_;
    std::string TEMPLATE_SURFACE;
    std::string surface_common_template_id_;
    std::string TEMPLATE_SEARCH;
    bool source_finder_;
};

#endif
