/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 *
 */

#include "query.h"
#include "scope.h"
#include "aggchildscope.h"
#include "listener.h"
#include "i18n.h"
#include "utils.h"

#include <locale.h>

#include <iostream>

#include <unity/scopes/Annotation.h>
#include <unity/scopes/CategorisedResult.h>
#include <unity/scopes/CategoryRenderer.h>
#include <unity/scopes/QueryBase.h>
#include <unity/scopes/SearchMetadata.h>
#include <unity/scopes/SearchReply.h>
#include <unity/scopes/ChildScope.h>
#include <unity/scopes/Scope.h>
#include <unity/UnityExceptions.h>

#include <QDir>
#include <QDebug>
#include <QDateTime>

namespace us = unity::scopes;
using namespace std;

Query::Query(us::CannedQuery const& query,
                               us::SearchMetadata const& metadata,
                               const std::string scope_dir,
                               const std::string cache_dir,
                               us::RegistryProxy const& registry,
                               AggScope * agg_scope
                               ) :
    us::SearchQueryBase(query, metadata),
    scope_dir_(scope_dir),
    cache_dir_(cache_dir),
    registry_(registry),
    agg_scope_(agg_scope)
{
    setlocale(LC_ALL, "");
    scope_id = agg_scope_->scope_id;
}

Query::~Query()
{
}

void Query::cancelled()
{
}

const static string MSG = R"( 
{
    "schema-version": 1,
    "template": {
        "category-layout": "grid",
        "card-size": "large",
        "card-layout": "horizontal",
        "non-interactive": "true"
    },
    "components": {
        "title": "title",
        "subtitle": "subtitle"
    }
}
)";

void Query::run(us::SearchReplyProxy const& upstream_reply)
{

    if (!registry_) {
        throw us::ConfigException("No registry available, cannot locate child scopes");
        return;
    }

    current_child_scopes = agg_scope_->child_scopes();

    std::string query_string = query().query_string();

    bool empty_search = query_string.empty();

    load_hints();

    qDebug() << "==== HINTS. uses_hints" << uses_hints;
    qDebug() << "==== HINTS. display_hints_on_firstuse" << hints_quickstart;
    qDebug() << "==== HINTS. query_string" << qstr(query_string);

    if (uses_hints && hints_quickstart && hints_exists() && query_string == HINTS_HIDE)
        dismiss_hints_quickstart();

    if (uses_hints && hints_quickstart && hints_exists() && !is_hints_quickstart_dismissed())
    {
        display_hints_quickstart(upstream_reply);
        return;
    }

    bool has_network = true;
    bool has_location = true;

    if (search_metadata().internet_connectivity() != us::QueryMetadata::ConnectivityStatus::Connected)
        has_network = false;
    if (!search_metadata().has_location())
        has_location = false;

    if (search_metadata().internet_connectivity() == us::QueryMetadata::ConnectivityStatus::Connected)
        qDebug() << "==== NET: connected";
    else if (search_metadata().internet_connectivity() == us::QueryMetadata::ConnectivityStatus::Disconnected)
        qDebug() << "==== NET: disconnected";
    else if (search_metadata().internet_connectivity() == us::QueryMetadata::ConnectivityStatus::Unknown)
        qDebug() << "==== NET: unknown";
    else
        qDebug() << "==== NET: (NONE)";

    qDebug() << "==== HINTS no_network_msg:" << no_network_msg;
    qDebug() << "==== HINTS no_network_msg_hints:" << no_network_msg_hints;
    qDebug() << "==== HINTS no_location_msg:" << no_location_msg;
    qDebug() << "==== HINTS no_location_msg_hints:" << no_location_msg_hints;
    if (!has_network)
    {
        qDebug() << "==== HINTS NO NET";
        if (no_network_msg)
        {
            qDebug() << "==== HINTS NO NET. simple message";
            std::string cat_id = "no_net_msg";
            us::Category::SCPtr msg_cat;
            if (!upstream_reply->lookup_category(cat_id))
		    msg_cat = upstream_reply->register_category(cat_id, "", "", us::CategoryRenderer(MSG));
            us::CategorisedResult res(msg_cat);
            res.set_uri("not_a_link");
            QString tdir = QString("%1/locale/").arg(qstr(scope_dir_));
            bindtextdomain(GETTEXT_DOMAIN_LIB.toStdString().c_str(), tdir.toStdString().c_str());
            std::string msg = dgettext(GETTEXT_DOMAIN_LIB.toStdString().c_str(), "Please connect to the internet. Refresh by pulling down the middle of the screen.");
            res.set_title("<b>" + msg + "</b>");
            upstream_reply->push(res);
            return;
        }
        else if (no_network_msg_hints && uses_hints && hints_exists())
        {
            qDebug() << "==== HINTS NO NET. message from hints";
            us::Category::SCPtr hint_cat;
            us::SearchMetadata metadata(search_metadata());

            auto reply = std::make_shared<ResultForwarder>(upstream_reply,
                [this, &hint_cat](us::CategorisedResult& res) -> bool {
                                        return true;
            });

            subsearch(hints_scope, "noconnection", HINTS_THIS_SCOPE, us::FilterState(),
                      metadata, reply);

            return;
        }
    }
    else  if (!has_location)
    {
        qDebug() << "==== HINTS: NO LOCATION";
        if (no_location_msg)
        {
            std::string cat_id = "no_location_msg";
            us::Category::SCPtr msg_cat;
            if (!upstream_reply->lookup_category(cat_id))
		    msg_cat = upstream_reply->register_category(cat_id, "", "", us::CategoryRenderer(MSG));

            QString tdir = QString("%1/locale/").arg(qstr(scope_dir_));
            bindtextdomain(GETTEXT_DOMAIN_LIB.toStdString().c_str(), tdir.toStdString().c_str());

            us::CategorisedResult res(msg_cat);
            res.set_uri("not_a_link");
            std::string msg = dgettext(GETTEXT_DOMAIN_LIB.toStdString().c_str(), "Please enable location services. Refresh by pulling down the middle of the screen");
            res.set_title("<b>" + msg + "</b>");
            upstream_reply->push(res);
            return;
        }
        else if (no_location_msg_hints && uses_hints && hints_exists())
        {
            qDebug() << "==== HINTS NO LOCATION. message from hints";
            us::Category::SCPtr hint_cat;
            us::SearchMetadata metadata(search_metadata());

            auto reply = std::make_shared<ResultForwarder>(upstream_reply,
                [this, &hint_cat](us::CategorisedResult& res) -> bool {
                                        return true;
            });

            subsearch(hints_scope, "nopositioning", HINTS_THIS_SCOPE, us::FilterState(),
                      metadata, reply);

            return;
        }
    }

    //load declared cardinality settings, if any
    load_cardinality_settings();

    //load scope and keyword declarations
    load_declarations();

    // create declared child scope instances and depts from child_scopes.json
    make_declared_scopes();

    // create child scope instances from keywords
    make_keyword_scopes();

    // create child scope instances from keywords and scopes
    make_categories();

    // complete initializations inlcuding cutting child scopes to those present and enabled
    handle_current_child_scopes(empty_search, upstream_reply);

    // convert current_scopes into scope_order based on declared order
    set_scope_order();

    //check that some sources are installed
    if (!has_one_source)
        noSources(upstream_reply);

    //optionally create departments from those declared and therefore in depts map
    if (using_departments)
    {
        create_departments(upstream_reply);
    }
    else
        qWarning () << QString("%1: NOT using departments").arg(qstr(SCOPE_ID));

    // set up handling for results and handle them, including buffering to ensure order
    handle_child_scope_results(query_string, empty_search, upstream_reply);

}

void Query::check_result_fallbacks(us::CategorisedResult &res, std::string common_id)
{
    qDebug() << "=== check FALLB common id: "<< qstr(common_id);
    for (auto fb : common_templates_fallbacks)
    {
        qDebug() << "=== check FALLB RES. a fallback found";
    }
    if ( common_templates_fallbacks.find(common_id) != common_templates_fallbacks.end())
    {
        qDebug() << "=== check FALLB common id found";
        auto map = *common_templates_fallbacks[common_id];
        for(auto iter = map.begin(); iter != map.end(); ++iter)
        {
            qDebug() << "==== check FALLB first: " << qstr(iter->first);
            for (auto field : iter->second)
            {
                qDebug() << "==== check FALLB field: " << qstr(field);
                if (res[iter->first].get_string().empty())
                {
                    qDebug() << "==== check FALLB first is empty: " << qstr(res[iter->first].get_string());
                    if (res.contains(field))
                        res[iter->first] = res[field];
                }
            }
        }
    }
}

/**
 * Cycle through scopes in order. For each, set up buffering and handle the current CategorisedResult (or use default handling). Results from the first scope are not buffered, that is they are displayed as soon as they arrive. Results from other scopes are held (buffered) until all preceding scopes have results, thus ensuring scope ordering
 */
void Query::handle_child_scope_results(std::string query_string, bool empty_search, us::SearchReplyProxy const& upstream_reply)
{
    std::vector<std::shared_ptr<ResultForwarder>> replies; // used for buffering/category ordering
    std::vector<std::string> added_by_localId;
    qDebug() << "==== scopes_ordered.size(): " << scopes_ordered.size();
    int idx = -1;//used in scopes_m_int_localId to store scope local_id
    for (const std::string &local_id : scopes_ordered)
    {
        // get a pointer to the current child scope (by id from the map)
        std::shared_ptr<AggChildScope> scope = scopes_m[local_id];

        us::CannedQuery canned_query = query();
        std::string dept_id = canned_query.department_id();

        // skip to next if scope is not registered in the system
        auto reg_scopes = registry_->list();
        us::MetadataMap::const_iterator scope_it;
        scope_it = reg_scopes.find(localId_id_m[local_id]);
        if (scope_it == reg_scopes.end()) {
            qWarning () << QString("%1: scope is NOT REGISTERED, skipping: %2").arg(qstr(SCOPE_ID), qstr(localId_id_m[local_id]));
            continue;
        }

        //TODO: add to docs note that when declaring a scope more than once, all must be in different depts
        if (using_departments)
        {
            if (dept_id == "") //is root dept
            {
                if(scope->department() != dept_id_of_root)
                   continue;
            }
            else if(scope->department() != query().department_id())
                continue;
        }
        
        // if scope is not in the currently enabled list, skip it
        bool found_as_enabled = false;
        for (auto cs : current_child_scopes)//FQ ids
        {
            if (cs.id == localId_id_m[local_id])
            {
                if (cs.enabled)
                {
                    found_as_enabled = true;
                }
            }
        }

        if (!found_as_enabled)
        {
            qWarning () << QString("%1: scope is NOT ENABLED, skipping: %2").arg(qstr(SCOPE_ID), qstr(localId_id_m[local_id]));
            continue;
        }

        // if scope is in exclude scopes list, skip it
        bool found_as_exclude = false;
        for (auto kw : keywords)
        {
            if (kw->id == scope->keyword()) {
                auto iter = std::find(kw->exclude_scopes.begin(), kw->exclude_scopes.end(), scope->id());
                if (iter != kw->exclude_scopes.end())
                {
                    found_as_exclude = true;
                    qWarning () << QString("==== FOUND AS EXCLUDE:  %1: scope is in exclude scopes of keyword: %2")
                                   .arg(qstr(scope->id()), qstr(scope->keyword()));
                }
            }
        }

        if (found_as_exclude) {
            continue;
        }


        if (query_string.empty() && scope->only_in_search() == true)
        {
            continue;
        }

        idx++;
        scopes_m_int_localId[idx] = local_id; // add id to map used later to get dept_id
        qDebug() << "==== scopes scopes_m_int_localId[idx] init: " << qstr(scopes_m_int_localId[idx]);

        // Add the scope (by its proxy) to our list of enabled scopes
        auto iter_ = find(added_by_localId.begin(), added_by_localId.end(), scope->local_id());
        if (iter_ == added_by_localId.end())
        {
            added_by_localId.emplace_back(scope->local_id());
            scopes.push_back(scope->proxy());
        }
        else
            continue;

        if (idx == 0)

            scope->set_buffer(false); // means results are displayed as they arrive
        else
            scope->set_buffer(true); // means results held (buffered) until all prevoius scopes have at least one result


        qDebug() << "==== scope init id: " << qstr(scope->id());
        qDebug() << "==== scope init local_id: " << qstr(scope->local_id());
        qDebug() << "==== scope init buffered" << scope->buffer();
        qDebug() << "==== scope init keyword" << scope->keyword_scope();
        qDebug() << "==== scope init shared cat name" << qstr(scope->keyword_scope_shared_cat_name());
        qDebug() << "==== scope init scopes.size(): " << scopes.size();

        // this lambda is run by ResultFowarder instance for each result
        // its return value determines whether the result needs to be buffered to ensure category ordering
        // results deemed invalid (for example of category that does not match a declared category) are marked by
        // addinng "dont_use" key and are filtered out in ResulttForwarded::push
        // all adjustments to the result are done here to match the child scopes declarations

        auto reply = make_shared<Listener>(upstream_reply,scope,
                    [canned_query, scope, this, &upstream_reply,empty_search,query_string](us::CategorisedResult& res) -> bool {

            qDebug() << "===== RES lambda.  is added: " << qstr(scope->local_id());
            qDebug() << "===== RES lambda. keyword: " << scope->keyword_scope();
            qDebug() << "===== RES lambda. shared_cat_name: " << qstr(scope->keyword_scope_shared_cat_name());
            qDebug() << "===== RES lambda. using depts: " << this->using_departments;
            if (this->using_departments)
            {
                if (canned_query.department_id() == "")//surfacing, show root dept only
                {
                    if (scope->department() != dept_id_of_root)
                    {
                        res["dont_use"] = "true";
                    }
                }
                else
                {
                    if (scope->department() != canned_query.department_id())
                        res["dont_use"] = "true";
                }
            }
            // save bits of incoming result
            std::string inc_res_cat_id = res.category()->id();
            us::Category::SCPtr res_cat = res.category();
            std::string inc_res_cat_title = res.category()->title();
            std::string inc_res_rdr = res.category()->renderer_template().data();

            qDebug() << "==== RESULT scope: " << qstr(scope->id());

            // for use in category registration
            std::string cat_id;
            std::string cat_title;
            std::string rdr;
            QString dateformat = "dd.MM.yyyy_hh:mm:ss:zzz"; // used as part of category ID to ensure uniqueness

            if (scope->keyword_scope())
            {
                std::string keyword = scope->keyword_scope_shared_cat_name();
                qDebug() << "==== KW RESULT is keyyword scope: " << qstr(scope->id());
                if (scope->keyword_uses_shared_cat())
                {
                    cat_title = _(shared_keyword_cat_titlemsgid[keyword].c_str());
                    qDebug() << "==== KW RESULT uses shared cat: " << qstr(keyword);

                    if (!query().query_string().empty())
                    {
                        qDebug()<<qstr("==== SEARCH scope's (%1) search template: %2").arg(qstr(scope->local_id()), qstr(keyword_catname_search_renderer[keyword]).replace("\n", " "));
                        if (keyword_catname_search_renderer.find(keyword) != keyword_catname_search_renderer.end()) // keyword has declared cat search renderer
                        {
                            qDebug() << "==== SEARCH keyword using search renderer";
                            cat_id = scope->local_id() + ":keyword:shared_cat:search_renderer::search";
                            rdr = keyword_catname_search_renderer[keyword];
                        }
                        else
                        {
                            qDebug() << "==== SEARCH using incoming result renderer";
                            cat_id = scope->local_id() + ":keyowrd:shared_cat:incoming_renderer::search";
                            rdr = inc_res_rdr;
                        }
                        if (!upstream_reply->lookup_category(cat_id))
                        {
                            catname_catptr[scope->keyword_scope_shared_cat_name()] = upstream_reply->register_category
                            (
                                cat_id,
                                cat_title,
                                 "",
                                us::CategoryRenderer(rdr)
                            );
                        }
                        res.set_category(catname_catptr[scope->keyword_scope_shared_cat_name()]);
                    }
                    else if ( keyword_catname_first_result_renderers.find(keyword) != keyword_catname_first_result_renderers.end() ) // uses first result rdr
                    {
                        if ((first_result_owner.empty()) || (first_result_owner == keyword))
                        {
                            qDebug() << "==== KW RESULT uses first result: " << qstr(scope->id());
                            if (keywordSharedCat_isFirstResult[keyword])
                            {
                                first_result_owner = keyword;
                                qDebug() << "==== KW RESULT is FIRST result: " << qstr(scope->keyword_scope_shared_cat_name());
                                qDebug() << "==== KW RESULT claims FIRST result: " << qstr(scope->keyword_scope_shared_cat_name());
                                cat_id = keyword + ":keyword:is_shared_category:is_first_result";
                                rdr = keyword_catname_first_result_renderers[keyword];
                                keywordSharedCat_isFirstResult[keyword] = false;
                                keywordSharedCat_isSecondResult[keyword] = true;
                                if (!upstream_reply->lookup_category(cat_id))
                                {
                                    catname_catptr[scope->keyword_scope_shared_cat_name()] = upstream_reply->register_category
                                    (
                                        cat_id,
                                        cat_title,
                                         "",
                                        us::CategoryRenderer(rdr)
                                    );
                                }
                            }
                            if ( keywordSharedCat_isSecondResult[keyword] )// second result
                            {
                                keywordSharedCat_isSecondResult[keyword] = false;
                                qDebug() << "==== KW RESULT uses FIRST resultm is SECOND result: " << qstr(scope->keyword_scope_shared_cat_name());
                                if (keyword_catname_renderer.find(keyword) != keyword_catname_renderer.end()) // keyword has declared cat renderer
                                {
                                    qDebug() << "==== KW RESULT uses FIRST result. is SECOND result. uses surface_template ";
                                    cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:uses_first_result:second_result:surface_template";
                                    rdr = keyword_catname_renderer[keyword];
                                }
                                else // use result template
                                {
                                    qDebug() << "==== KW RESULT uses FIRST result. is SECOND result. NOT uses serface_template ";
                                    cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:uses_first_result:second_result:result_template";
                                    rdr = res.category()->renderer_template().data();
                                }
                                if (!upstream_reply->lookup_category(cat_id))
                                {
                                    catname_catptr[scope->keyword_scope_shared_cat_name()] = upstream_reply->register_category
                                    (
                                        cat_id,
                                        cat_title,
                                         "",
                                        us::CategoryRenderer(rdr)
                                    );

                                }
                            }
                        }
                        else  // has declared first result but it is claimed by different keyword or scoope
                        {
                            if (keyword_catname_renderer.find(keyword) != keyword_catname_renderer.end()) // keyword has declared cat renderer
                            {
                                qDebug() << "==== KW RESULT uses FIRST result. other claims it.. uses surface_template ";
                                if (catname_catptr.find(keyword) == catname_catptr.end())
                                {
                                    cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:uses_first_result:other_claims_it:surface_template";
                                    rdr = keyword_catname_renderer[keyword];
                                    if (!upstream_reply->lookup_category(cat_id))
                                    {
                                        catname_catptr[keyword] = upstream_reply->register_category
                                        (
                                            cat_id,
                                            cat_title,
                                             "",
                                            us::CategoryRenderer(rdr)
                                        );
                                    }
                                }
                            }
                            else // use result template
                            {
                                qDebug() << "==== KW RESULT uses FIRST result. other claims it. uses result template ";
                                if (catname_catptr.find(keyword) == catname_catptr.end())
                                {
                                    cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:uses_first_result:other_claims_it:result_template";
                                    rdr = res.category()->renderer_template().data();
                                    if (!upstream_reply->lookup_category(cat_id))
                                    {
                                        catname_catptr[keyword] = upstream_reply->register_category
                                        (
                                            cat_id,
                                            cat_title,
                                             "",
                                            us::CategoryRenderer(rdr)
                                        );
                                    }
                                }
                            }
                        }
                        res.set_category(catname_catptr[scope->keyword_scope_shared_cat_name()]);
                    }
                    else // not search and does not use declared first result rdr
                    {
                        if (!scope->surface_template().empty())// has a decared surface template
                        {
                            qDebug() << "==== KW RESULT not uses first result. uses surface template";
                            cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:not_uses_first_result:surface_template";
                            qDebug() << "==== KW RESULT not uses first result. uses surface template. surface temlate: " << qstr(scope->surface_template()).replace("\n", " ");
                            rdr = scope->surface_template();
                        }
                        else // use result template
                        {
                            qDebug() << "==== KW RESULT not uses first result. NOT uses surface template";
                            cat_id = scope->keyword_scope_shared_cat_name() + ":keyword:is_shared_category:not_uses_first_result:result_template";
                            rdr = res.category()->renderer_template().data();
                        }
                        if (catname_catptr.find(keyword) == catname_catptr.end())
                        {
                            if (categoryId_linkToChildSpecified.find(scope->category_id()) != categoryId_linkToChildSpecified.end())
                            {
                                if (!upstream_reply->lookup_category(cat_id))
                                {
                                    scope->set_category
                                    (
                                        upstream_reply->register_category
                                        (
                                            cat_id,
                                            cat_title,
                                             "",
                                            us::CannedQuery(categoryId_linkToChildSpecified[scope->category_id()]),
                                            us::CategoryRenderer(rdr)
                                        )
                                    );
                                }
                            }
                            else if (!upstream_reply->lookup_category(cat_id))
                            {
                                catname_catptr[scope->keyword_scope_shared_cat_name()] = upstream_reply->register_category
                                (
                                    cat_id,
                                    cat_title,
                                     "",
                                    us::CategoryRenderer(rdr)
                                );

                            }
                        }
                        res.set_category(catname_catptr[scope->keyword_scope_shared_cat_name()]);
                    }
                }
                else // keyword scope does not use a shared category
                {
                    if (scope->is_first_result())
                    {
                        qDebug() << "==== FIRST RESULT for scope: " << qstr(scope->local_id());
                        scope->set_is_first_result(false);
                        cat_title = inc_res_cat_title;
                        if (cat_title.empty()){
                            cat_title = registry_->get_metadata(scope->id()).display_name();
                        }

                        qDebug() << "==== RES. KEYWORD. local_id: " << qstr(scope->local_id());
                        qDebug() << "==== RES. KEYWORD. cat_ttile: " << qstr(cat_title);
                        if (keyword_catname_renderer.find(keyword) != keyword_catname_renderer.end()) // keyword has declared cat renderer
                        {
                            qDebug() << "==== KW RESULT not shared declared renderer ";
                            cat_id = scope->keyword() + ":keyword:not_shared_category:declared_renderer";
                            rdr = keyword_catname_renderer[keyword];
                        }
                        else // use result template
                        {
                            qDebug() << "==== KW RESULT not shared incoming renderer";
                            cat_id = scope->keyword() + ":keyword:not_shared_category:incoming_renderer";
                            rdr = res.category()->renderer_template().data();
                        }
                        if (!upstream_reply->lookup_category(cat_id))
                        {
                           if (scope->category_link_to_child()) 
                           {
                               scope->set_category
                               (
                                   upstream_reply->register_category
                                   (
                                    cat_id,
                                    cat_title,
                                    "",
                                    us::CannedQuery(scope->id(), query_string, scope->child_department()),
                                    us::CategoryRenderer(rdr)
                                   ) 
                               );
                           }
                           else
                           {
                               scope->set_category
                               (
                                   upstream_reply->register_category
                                   (
                                    cat_id,
                                    cat_title,
                                    "",
                                    us::CategoryRenderer(rdr)
                                   )
                               );
                           }
                        }
                    }
                    res.set_category(scope->category());
                }
            }
            else if (scope->category_scope())
            {
                qDebug() << "==== LINK loop categoryId_linkToChildSpecified size: " << categoryId_linkToChildSpecified.size();
                for (auto spec : categoryId_linkToChildSpecified)
                    qDebug() << "==== LINK loop: " << qstr(spec.second);

                std::string category_id = scope->category_id();
                qDebug() << "==== CATREG RESULT scope: " << qstr(scope->id());

                if (!categoryId_titleMsgid[category_id].empty())
                    cat_title = _(categoryId_titleMsgid[category_id].c_str());
                else
                    cat_title = "";
                qDebug() << "==== CATREG RESULT category id: " << qstr(category_id);
                qDebug() << "==== CATREG RESULT category title msid: " << qstr(cat_title);

                if (!query().query_string().empty())
                {
                    qDebug()<<qstr("==== CATREG scope's (%1) search template: %2").arg(qstr(scope->local_id()), qstr(scope->search_template()).replace("\n", " "));
                    if (!scope->search_template().empty())
                    {
                        qDebug() << "==== CATREG keyword using search renderer";
                        cat_id = scope->local_id() + ":category:search_renderer::search";
                        rdr = scope->search_template();
                    }
                    else
                    {
                        qDebug() << "==== CATREG using incoming result renderer";
                        cat_id = scope->local_id() + ":category:incoming_renderer::search";
                        rdr = inc_res_rdr;
                    }
                    if (!upstream_reply->lookup_category(cat_id))
                    {
                        catname_catptr[scope->keyword_scope_shared_cat_name()] = upstream_reply->register_category
                        (
                            cat_id,
                            cat_title,
                             "",
                            us::CategoryRenderer(rdr)
                        );
                    }
                    res.set_category(catname_catptr[scope->keyword_scope_shared_cat_name()]);
                }
                else if ( categoryId_first_result_renderers.find(category_id) != categoryId_first_result_renderers.end() ) // uses first result rdr
                {
                    qDebug() << "==== CATREG RESULT FIRST RESULT: " << qstr(category_id);

                    if ((first_result_owner.empty()) || (first_result_owner == category_id))
                    {
                        // perform fallbacks, if any
                        if(!scope->first_common_template_id().empty())
                        {
                            qDebug() << "==== first check FALLB. scope->first_common_template_id():" << qstr(scope->first_common_template_id());
                           check_result_fallbacks(res, scope->first_common_template_id());
                        }
                        qDebug() << "==== CATREG RESULT uses first result: " << qstr(scope->id());
                        if (categoryId_isFirstResult[category_id])
                        {
                            first_result_owner = category_id;
                            qDebug() << "==== CATREG RESULT is FIRST result: " << qstr(scope->category_id());
                            qDebug() << "==== CATREG RESULT claims FIRST result: " << qstr(scope->category_id());
                            cat_id = category_id + ":category:is_first_result";
                            rdr = categoryId_first_result_renderers[category_id];
                            categoryId_isFirstResult[category_id] = false;
                            categoryId_isSecondResult[category_id] = true;
                            if (categoryId_linkToChildSpecified.find(scope->category_id()) != categoryId_linkToChildSpecified.end())
                            {
                                if (!upstream_reply->lookup_category(cat_id))
                                {
                                    catname_catptr[category_id] = upstream_reply->register_category
                                    (
                                        cat_id,
                                        cat_title,
                                         "",
                                        us::CannedQuery(categoryId_linkToChildSpecified[scope->category_id()]),
                                        us::CategoryRenderer(rdr)
                                    );
                                }
                            }
                            else if (!upstream_reply->lookup_category(cat_id))
                            {
                                catname_catptr[category_id] = upstream_reply->register_category
                                (
                                    cat_id,
                                    cat_title,
                                     "",
                                    us::CategoryRenderer(rdr)
                                );
                            }
                        }
                        else if ( categoryId_isSecondResult[category_id] )// second result
                        {
                            categoryId_isSecondResult[category_id] = false;
                            qDebug() << "==== CATREG RESULT uses FIRST resultm is SECOND result: " << qstr(category_id);
                            // perform fallbacks, if any
                            if(!scope->surface_common_template_id().empty())
                            {
                                qDebug() << "==== surface check FALLB. scope->surface_common_template_id():" << qstr(scope->surface_common_template_id());
                               check_result_fallbacks(res, scope->surface_common_template_id());
                            }
                            qDebug()<<"==== CATREG SURFACE: " << qstr(scope->surface_template()).replace("\n"," ");
                            if (!scope->surface_template().empty()) // keyword has declared cat renderer
                            {
                                qDebug() << "==== CATREG R=ESULT uses FIRST result. is SECOND result. uses surface_renderer ";
                                cat_id = category_id + ":category:uses_first_result:second_result:surface_template";
                                rdr = scope->surface_template();
                            }
                            else // use result template
                            {
                                qDebug() << "==== CATREG RESULT uses FIRST result. is SECOND result. NOT uses surface_template ";
                                cat_id = category_id + ":category:uses_first_result:second_result:result_template";
                                rdr = res.category()->renderer_template().data();
                            }
                            if (!upstream_reply->lookup_category(cat_id))
                            {
                                catname_catptr[category_id] = upstream_reply->register_category
                                (
                                    cat_id,
                                    cat_title,
                                     "",
                                    us::CategoryRenderer(rdr)
                                );

                            }
                        }
                        else  // has declared first result but it is claimed by different keyword or scoope
                        {
                            // perform fallbacks, if any
                            if(!scope->surface_common_template_id().empty())
                            {
                                qDebug() << "==== surface check FALLB. scope->surface_common_template_id():" << qstr(scope->surface_common_template_id());
                               check_result_fallbacks(res, scope->surface_common_template_id());
                            }
                            if (categoryId_renderer.find(category_id) != categoryId_renderer.end()) // keyword has declared cat renderer
                            {
                                qDebug() << "==== CATREG RESULT uses FIRST result. other claims it. using surface_template ";
                                if (catname_catptr.find(category_id) == catname_catptr.end())
                                {
                                    cat_id = category_id + ":category:uses_first_result:other_claims_it:surface_template";
                                    rdr = categoryId_renderer[category_id];
                                    if (!upstream_reply->lookup_category(cat_id))
                                    {
                                        catname_catptr[category_id] = upstream_reply->register_category
                                        (
                                            cat_id,
                                            cat_title,
                                             "",
                                            us::CategoryRenderer(rdr)
                                        );
                                    }
                                }
                            }
                            else // use result template
                            {
                                qDebug() << "==== CATREG RESULT uses FIRST result. other claims it. uses result template ";
                                if (catname_catptr.find(category_id) == catname_catptr.end())
                                {
                                    cat_id = category_id + ":category:uses_first_result:other_claims_it:result_template";
                                    rdr = res.category()->renderer_template().data();
                                    if (!upstream_reply->lookup_category(cat_id))
                                    {
                                        catname_catptr[category_id] = upstream_reply->register_category
                                        (
                                            cat_id,
                                            cat_title,
                                             "",
                                            us::CategoryRenderer(rdr)
                                        );
                                    }
                                }
                            }
                        }
                        res.set_category(catname_catptr[category_id]);
                    }
                }
                else // not search and does not use declared first result rdr
                {
                    // if uses common template, perform fallbacks, if any
                    if(!scope->surface_common_template_id().empty())
                    {
                        qDebug() << "==== surface check FALLB. scope->surface_common_template_id():" << qstr(scope->surface_common_template_id());
                       check_result_fallbacks(res, scope->surface_common_template_id());
                    }


                    if (!scope->surface_template().empty())// has a decared surface template
                    {
                        qDebug() << "==== CATREG RESULT not uses first result. uses surface template";
                        cat_id = scope->category_id() + ":category:not_uses_first_result:surface_template";
                        qDebug() << "==== CATREG RESULT not uses first result. uses surface template. surface temlate: " << qstr(scope->surface_template()).replace("\n", " ");
                        rdr = scope->surface_template();
                    }
                    else // use result template
                    {
                        qDebug() << "==== CATREG RESULT not uses first result. NOT uses surface template";
                        qDebug() << "==== CATREG RESULT not uses first result. does use incoming result (not impld yet): " << qstr(category_id);
                        cat_id = scope->keyword_scope_shared_cat_name() + ":category:not_uses_first_result:result_template";
                        rdr = res.category()->renderer_template().data();
                    }
                    if (catname_catptr.find(category_id) == catname_catptr.end())
                    {
                        if (categoryId_linkToChildSpecified.find(scope->category_id()) != categoryId_linkToChildSpecified.end())
                        {
                            if (!upstream_reply->lookup_category(cat_id))
                            {
                                catname_catptr[category_id] = upstream_reply->register_category
                                (
                                    cat_id,
                                    cat_title,
                                     "",
                                    us::CannedQuery(categoryId_linkToChildSpecified[scope->category_id()]),
                                    us::CategoryRenderer(rdr)
                                );
                            }
                        }
                        else if (!upstream_reply->lookup_category(cat_id))
                        {
                            catname_catptr[scope->category_id()] = upstream_reply->register_category
                            (
                                cat_id,
                                cat_title,
                                 "",
                                us::CategoryRenderer(rdr)
                            );

                        }
                    }
                    res.set_category(catname_catptr[category_id]);
                }
            }
            else // is a declared scope
            {
                qDebug() << "==== RES. scope is declared: " << qstr(scope->local_id());
                bool set_declared_category = false;


                // if child_category set, filter out others by adding "dont_use" key to result
                // also impose max category results if set
                if (!scope->child_category().empty())
                {
                    if (qstr(inc_res_cat_id) != qstr(scope->child_category()))
                    {
                        res["dont_use"] = "true";
                    }
                    else if (scope->using_child_category_max_results())
                    {
                        scope->inc_result_idx(); // allows constraining num results to child_category_max_results declaration
                        if (scope->result_idx() > scope->child_category_max_results() )
                        {
                            res["dont_use"] = "true";
                        }
                    }
                }

                if (scope->is_first_result())
                {
                    qDebug() << "==== FIRST RESULT for scope: " << qstr(scope->local_id());
                    scope->set_is_first_result(false);
                    scope->set_is_second_result(true);
                    set_declared_category = true;
                    // set catogery title when using title from first result.
                    // TODO: consider allowing declaration to control whether link to child used
                    QString dateformat = "dd.MM.yyyy_hh:mm:ss:zzz"; // used as part of category ID to ensure uniqueness

                    if (scope->using_category_title_incoming() )
                    {
                        scope->set_category_title(inc_res_cat_title);
                    }
                    else if (scope->using_category_title_display_name() )
                    {
                        scope->set_category_title(scope->metadata()->display_name());
                    }

                    cat_title = scope->category_title();

                    cat_id = scope->id() + ":declared_default";

                    qDebug() << "==== first_result_owner: " << qstr(first_result_owner);

                    // dont use first result renderer when searching
                    if (!query().query_string().empty())
                    {
                        if (!scope->search_template().empty())
                        {
                            qDebug() << "==== FIRST RESULT: SEARCH using search renderer";
                            cat_id = scope->local_id() + ":declared:first_result:search_renderer::search";
                            rdr = scope->search_template();
                        }
                        else if (!scope->surface_template().empty())
                        {
                            qDebug() << "==== FIRST RESULT: SEARCH using surface renderer";
                            cat_id = scope->local_id() + ":declared:first_result:surface_renderer::search";
                            rdr = scope->surface_template();
                        }
                        else
                        {
                            qDebug() << "==== FIRST RESULT: SEARCH using incoming result renderer";
                            cat_id = scope->local_id() + ":declared:first_result:incomiing_renderer::search";
                            rdr = inc_res_rdr;
                        }
                    }
                    else if (first_result_owner.empty() || first_result_owner == scope->id())
                    {
                        first_result_owner = scope->id();
                        if (!scope->first_result_template().empty())
                        {
                            qDebug() << "==== FIRST RESULT: using first result template";
                            cat_id = scope->local_id() + ":declared:first_result:first_result_template::surface";
                            rdr = scope->first_result_template();
                        }
                        else if (!scope->surface_template().empty())
                        {
                            cat_id = scope->local_id() + ":declared:first_result:surface_template::surface";
                            rdr = scope->surface_template();
                            qDebug() << "==== FIRST RESULT: using surface template";
                        }
                        else
                        {
                            cat_id = scope->local_id() + ":declared:first_result:incoming_template::surface";
                            rdr = inc_res_rdr;
                            qDebug() << "==== FIRST RESULT: using incoming template";
                        }
                    }
                    else if (!scope->surface_template().empty())
                    {
                        cat_id = scope->local_id() + ":declared:first_result:surface_template::surface";
                        rdr = scope->surface_template();
                        qDebug() << "==== FIRST RESULT: using surface template";
                    }
                    else
                    {
                        cat_id = scope->local_id() + ":declared:first_result:incoming_template::surface";
                        rdr = inc_res_rdr;
                        qDebug() << "==== FIRST RESULT: using incoming template";
                    }

                }
                else if
                (
                    scope->is_second_result()
                    && !scope->first_result_template().empty()
                    && !query().query_string().empty()
                )
                {
                    qDebug() << "==== SECOND RESULT for scope: " << qstr(scope->local_id());
                    scope->set_is_first_result(false);
                    scope->set_is_second_result(false);
                    set_declared_category = true;

                    if (scope->using_category_title_incoming() )
                    {
                        scope->set_category_title(inc_res_cat_title);
                    }
                    else if (scope->using_category_title_display_name() )
                    {
                        scope->set_category_title(scope->metadata()->display_name());
                    }

                    cat_title = scope->category_title();

                    QString dateformat = "dd.MM.yyyy_hh:mm:ss:zzz"; // used as part of category ID to ensure uniqueness
                    if (scope->category_link_to_child())
                    {
                        if (empty_search) // surfacing
                        {
                            qDebug() << " CAT: scope->category_title(): " << qstr(scope->category_title());
                            if (scope->override_surface_template()) // use declared renderer
                            {
                                cat_id = scope->local_id() + ":declared:not_first_result:link-override:surface_template:surface";
                                rdr = scope->surface_template();
                            }
                            else // use incoming renderer
                            {
                                cat_id = scope->local_id() + ":declared:not_first_result:link-override:incoming_template:surface";
                                rdr = inc_res_rdr;
                            }
                        }
                        else // user is searching
                        {
                            if (scope->override_search_template())
                            {
                                cat_id = scope->local_id() + ":declared:not_first_result:link-override:search_template:search";
                                rdr = scope->search_template().data();
                            }
                            else
                            {
                                cat_id = scope->local_id() + ":declared:not_first_result:link-override:incoming_template:search";
                                rdr = inc_res_rdr;
                            }
                        }
                    }
                    else // no category link to child
                    {
                        if (empty_search)
                        {
                            if (scope->override_surface_template())
                            {
                                cat_id = scope->local_id() + "_nolink-override-surface";
                                rdr = scope->surface_template().data();
                            }
                            else if (!scope->first_result_template().empty())
                            {
                                cat_id = scope->local_id() + "_nolink-nooverride-surface";
                                rdr = inc_res_rdr;;
                            }
                        }
                        else
                        {
                            if (scope->override_search_template())
                            {
                                cat_id = scope->local_id() + "_nolink-override-search";
                                rdr = scope->search_template().data();
                            }
                            else
                            {
                                cat_id = scope->local_id() + "_nolink-nooverride-search";
                                rdr = inc_res_rdr;;
                            }
                        }
                    }
                }
                if (set_declared_category)
                {
                    if (!rdr.empty()) // should not ever be empty!
                    {
                        if (!upstream_reply->lookup_category(cat_id))
                        {
                            if (scope->category_link_to_child())
                            {
                                scope->set_category
                                (
                                   upstream_reply->register_category
                                   (
                                    cat_id,
                                    cat_title,
                                    "",
                                    us::CannedQuery(scope->id(), query_string, scope->child_department()),
                                    us::CategoryRenderer(rdr)
                                   )
                                );
                            } 
                            else
                            {
                               scope->set_category
                               (
                                   upstream_reply->register_category
                                   (
                                    cat_id,
                                    cat_title,
                                    "",
                                    us::CategoryRenderer(rdr)
                                   )
                               );
                           }
                       }
                    }
                    else
                        qWarning() << QString("Warning: Scope %1 renderer is empty.").arg(qstr(scope->id()));
                }

                //for child scopes that declare attribute swaps, do them
                for (std::shared_ptr<std::pair<std::string,std::string>> swap : this->child_scopes_m[scope->local_id()]->swap_result_attributes)
                {
                    qDebug() << "==== trying to swap attributes";
                    try
                    {
                        if (!res[swap->first].get_string().empty())
                        {
                            qDebug() << "==== swap first: " << qstr(swap->first);
                            res[swap->second] = res[swap->first];
                        }
                    }
                    catch (unity::LogicException)
                    {
                        qDebug() << "==== failed to swap attributes";
                        qWarning() << QString("Scope %1 result does not contain attribute 2").arg(qstr(scope->local_id()), qstr(swap->second));
                    }
                }
                if (common_templates.find(child_scopes_m[scope->local_id()]->result_category_id_to_common_template[res.category()->id()]) != common_templates.end())
                {
                    std::string common_template_id = child_scopes_m[scope->local_id()]->result_category_id_to_common_template[res.category()->id()];
                    std::string cat_id = common_template_id + ":" + scope->local_id();

                    // if the common category has not yet been registered, register it
                    if (!upstream_reply->lookup_category(cat_id))
                    {
                        std::string template_to_use = common_templates[child_scopes_m[scope->local_id()]->result_category_id_to_common_template[res.category()->id()]];
                        if (scope->category_link_to_child())
                        {
                            registered_common_categories[cat_id] = upstream_reply->register_category(cat_id, scope->category_title(), "",
                                                            us::CannedQuery(scope->id(), query_string, scope->child_department()),
                                                       us::CategoryRenderer(template_to_use));
                        }
                        else
                        {
                            registered_common_categories[cat_id] = upstream_reply->register_category(cat_id, scope->category_title(), "",
                                                       us::CategoryRenderer(template_to_use));
                        }
                    }
                    res.set_category(registered_common_categories[cat_id]);
                }
                else //set result category
                    res.set_category(scope->category());
            }
            qDebug() << "==== RES CAT: " << qstr(res.category()->id());
            return true;
        });
        replies.push_back(reply);
    }

    qDebug() << "==== scopes.size(): " << scopes.size();
    qDebug() << "==== replies.size(): " << replies.size();
    // create and chain result forwarders to enforce proper order of categores
    for (unsigned int i = 1; i < scopes.size(); ++i) {
        for (unsigned int j = 0; j < i; ++j) {
            replies[j]->add_observer(replies[i]);
        }
    }

    // dispatch search to subscopes
    // also set cardinality, default and per scope
    for (unsigned int i = 0; i < replies.size(); ++i) {
        us::SearchMetadata metadata(search_metadata());
        auto scope = scopes_m[scopes_m_int_localId[i]];

        if (scope->keyword_scope())
        {
            if (id_keyword_map[scope->keyword()]->cardinality > -1)
            {
                metadata.set_cardinality(id_keyword_map[scope->keyword()]->cardinality);
            }
            else
            {
                metadata.set_cardinality(keyword_cardinality);
            }
        }
        else if (scope->category_scope())
        {
            if (id_category_m[scope->category_id()]->cardinality > -1)
            {
                qDebug()<< "=== SUBS is cat scope, card: " << id_category_m[scope->category_id()]->cardinality;
                metadata.set_cardinality(id_category_m[scope->category_id()]->cardinality);
            }
            else
            {
                metadata.set_cardinality(keyword_cardinality);
            }
        }
        else if (empty_search && !scope->using_child_category_max_results()) {
            metadata.set_cardinality( scope->cardinality() );
        }


        qDebug() << "===- LOCAL_ID: " << qstr(scope->local_id());
        metadata.set_hint(scope->local_id(), us::Variant("valueMeansNothing"));

        std::string query_text = query_string;

        // if scope is clickstore, set query string to open to scope
        if ( scope->source_finder())
        {
            if (query_store_ != "")
                query_text = query_store_;
            else
                query_text = query_string;
        }

        // dispatch subsearches differently depending on whether is keyword or declared scope
        if (scopes_m_int_localId[i] == "com.canonical.scopes.clickstore")
        {
            for (auto ch : current_child_scopes)
            {
                if (ch.id == localId_id_m[scopes_m_int_localId[i]])
                {
                    qDebug() << "=== subsearch CLICKSTORE scope: " << qstr(scopes_m_int_localId[i]);
                    subsearch
                    (
                        ch,
                        query_store_,
                        "",
                        us::FilterState(),
                        metadata,
                        replies[i]
                    );
                    break;
                }

            }
        }
        else if (scope->keyword_scope())
        {
            for (auto ch : current_child_scopes)
            {
                if (ch.id == localId_id_m[scopes_m_int_localId[i]])
                {
                    qDebug() << "=== subsearch KEYWORD keyword scope: " << qstr(scopes_m_int_localId[i]);
                    std::set<std::string> keywords;
                    keywords.insert(scope->keyword());
                    ch.keywords = keywords;
                    for ( auto ky : ch.keywords )
                        qDebug() << "=== subsearch KEYWORD with keyword: " << qstr(ky);
                    subsearch
                    (
                        ch,
                        query_string,
                        "",
                        us::FilterState(),
                        metadata,
                        replies[i]
                    );
                    break;
                }

            }
        }
        else
        {
            for (auto ch : current_child_scopes)
            {

                if (ch.id == localId_id_m[scopes_m_int_localId[i]])
                {
                    qDebug() << "=== subsearch non KEYWORD scope local_ id " << qstr(scope->local_id());
                    qDebug() << "=== subsearch non KEYWORD scope id " << qstr(scope->id());
                    qDebug() << "=== subsearch non KEYWORD keyword" << qstr(scope->keyword());
                    if (!scope->keyword().empty())
                    {
                        std::set<std::string> keywords;
                        keywords.insert(scope->keyword());
                        ch.keywords = keywords;
                        for ( auto ky : ch.keywords )
                            qDebug() << "=== subsearch KEYWORD with keyword: " << qstr(ky);
                    }
                    // when query string is empty, use scopes default_query_string, if any
                    std::string adj_qry;
                    if (!query_string.empty())
                        adj_qry = query_string;
                    else if (!localId_defaultQryStr[scopes_m_int_localId[i]].empty())
                        adj_qry = localId_defaultQryStr[scopes_m_int_localId[i]];
                    else
                        adj_qry = "";

                    subsearch
                    (
                        ch,
                        adj_qry,
                        scope->child_department(),
                        us::FilterState(),
                        metadata, replies[i]
                    );
                    break;
                }
            }
        }
    }
}
