/*
 * Copyright (C) 2014 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 */

#include "scope.h"
#include "query.h"
#include "preview.h"
#include "i18n.h"
#include "utils.h"

#include <unity-scopes.h>

#include <QDebug>
#include <QFile>
#include <QFile>
#include <QJsonParseError>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>

namespace us = unity::scopes;

AggScope::AggScope()
{
}

void AggScope::start(std::string const& scope_name)
{
    qDebug() <<"==== AggScope scope_name" << qstr(scope_name);
    scope_id = scope_name;
}

void AggScope::stop()
{
}

us::SearchQueryBase::UPtr AggScope::search(us::CannedQuery const &q,
                                        us::SearchMetadata const &metadata)
{

    const QString scopePath = QString::fromStdString(scope_directory());
    const QString cachePath = QString::fromStdString(cache_directory());

    textdomain(GETTEXT_DOMAIN.toStdString().c_str());

    QString tdir = QString("%1/locale/").arg(scopePath);
    bindtextdomain(GETTEXT_DOMAIN.toStdString().c_str(), tdir.toStdString().c_str());

    us::SearchQueryBase::UPtr query(new Query(q, metadata,
                                                scope_directory(),
                                                cache_directory(),
                                                registry(),
                                                this
                                                ));
    return query;
}

us::PreviewQueryBase::UPtr AggScope::preview(us::Result const& result, us::ActionMetadata const& metadata) {
    us::PreviewQueryBase::UPtr preview(new Preview(result, metadata));
    return preview;
}

std::vector<std::string> AggScope::get_declared_scopes(IdType type) const
{
    std::vector<std::string> declared_scopes;

    QString f_fn_qs = QString("%1/child_scopes.json").arg(QString::fromStdString(scope_directory()));
    QFile f_f;
    f_f.setFileName(f_fn_qs);

    if (!f_f.exists())
    {
        qWarning() << QString("%1 not found. No settingsfor declared child scopes.").arg(f_fn_qs);
        return declared_scopes;
    }

    f_f.open(QIODevice::ReadOnly);
    QByteArray data = f_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);
    //qDebug() << doc.toJson();

    if (err.error != QJsonParseError::NoError)
    {
        qWarning() << QString("Failed to parse %1 data: ").arg(f_fn_qs) << err.errorString();
        return declared_scopes;
    }

    QJsonObject doc_o = doc.object();

    QJsonArray scopes_and_keywords_a  = doc_o["order"].toArray();

    for (const auto item_ : scopes_and_keywords_a)
    {
        QJsonObject item_o = item_.toObject();
        if (item_o.contains("scope"))
        {
            QJsonObject scope_o = item_o["scope"].toObject();
            if (type == IdType::local_id)
            {
                std::string local_id = scope_o["local_id"].toString().toStdString();
                declared_scopes.emplace_back(local_id);
            }
            else
            {
                std::string id = scope_o["id"].toString().toStdString();
                declared_scopes.emplace_back(id);
            }
        }
        else if (item_o.contains("category"))
        {
    qDebug() << "==== GET category found";
            QJsonObject cat_o = item_o["category"].toObject();
            if (cat_o.contains("scopes"))
            {
    qDebug() << "==== GET scopes found";
                QJsonArray scopes_a = cat_o["scopes"].toArray();
                for (auto item : scopes_a)
                {
    qDebug() << "==== GET scope found";
                    QJsonObject scope_o = item.toObject();
                    if (type == IdType::local_id)
                    {
                        std::string local_id = scope_o["local_id"].toString().toStdString();
                        declared_scopes.emplace_back(local_id);
                    }
                    else
                    {
    qDebug() << "==== GET id found";
                        std::string id = scope_o["id"].toString().toStdString();
                        declared_scopes.emplace_back(id);
                    }
                }
            }
        }
    }

    f_f.flush();
    f_f.close();
    return declared_scopes;
}

std::vector<std::string> AggScope::get_declared_keywords() const
{
    std::vector<std::string> keywords;

    QString f_fn_qs = QString("%1/child_scopes.json").arg(QString::fromStdString(scope_directory()));
    QFile f_f;
    f_f.setFileName(f_fn_qs);

    if (!f_f.exists())
    {
        qWarning() << QString("%1: child_scopes.json not found. Not aggregating keyworded child scopes.").arg(qstr(SCOPE_ID));
        return keywords;
    }

    f_f.open(QIODevice::ReadOnly);
    QByteArray data = f_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);

    if (err.error != QJsonParseError::NoError)
    {
        qWarning() << QString("%1:Failed to parse keywords.json data. Error: ").arg(qstr(SCOPE_ID), err.errorString());
        f_f.flush();
        f_f.close();
        return keywords;
    }

    QJsonObject doc_o = doc.object();

    QJsonArray order_a = doc_o["order"].toArray();

    for (const auto order_item_o : order_a)
    {
        QJsonObject o_ = order_item_o.toObject();

        if (o_.contains("keyword"))
        {
            QJsonObject keyword_o = o_["keyword"].toObject();
            keywords.emplace_back(keyword_o["keyword"].toString().toStdString());
        }
        if (o_.contains("category"))
        {
            QJsonObject category_o = o_["category"].toObject();
            if (category_o.contains("keywords"))
            {
                QJsonArray keywords_a = category_o["keywords"].toArray();
                for (const auto item_o : keywords_a)
                {
                    QJsonObject keyword_o = item_o.toObject();
                    std::string id = keyword_o["id"].toString().toStdString();
                    auto iter = std::find(keywords.begin(), keywords.end(), id);
                    if (iter == keywords.end())
                    {
                        keywords.emplace_back(id);
                        qDebug() << "==== FOUND: " << qstr(id);
                    }
                }
            }
        }
    }
    return keywords;
}


us::ChildScopeList AggScope::find_child_scopes() const
{
    std::vector<std::string> declared_scopes = get_declared_scopes(IdType::id);

    us::ChildScopeList list;
    std::map<std::string,us::ScopeMetadata> declared_registered_scopes;

    // ask registry for the declared scopes that are installed
    declared_registered_scopes = registry()->list_if([declared_scopes](us::ScopeMetadata const& md)
    {
        auto iter = find(declared_scopes.begin(), declared_scopes.end(), md.scope_id());
        if (iter != declared_scopes.end())
            return true;
    });

    for (auto mmap : declared_registered_scopes)
    {
        list.emplace_back
        (
            us::ChildScope
            {
                mmap.first, mmap.second, true
            }
        );
    }

    std::vector<std::string> keywords_l = get_declared_keywords();

    std::map<std::string,std::set<std::string>> map;
    auto kw_scopes = registry()->list_if([keywords_l, &map](us::ScopeMetadata const& item)
    {
        auto keywords = item.keywords();
        bool found = false;
        for (std::string kw : keywords_l)
        {
            if (keywords.find(kw) != keywords.end())
            {
                found = true;
                if (map.find(item.scope_id()) != map.end())
                {
                    map[item.scope_id()].emplace(kw);
                }
                else
                {
                    std::set<std::string> kws_set;
                    kws_set.emplace(kw);
                    map[item.scope_id()] = kws_set;
                }
            }
        }
        return found;//seems to be ok to return the same scope multiple times (in the case of multiple keyword matches)
    });

    for (auto const& kw_scope : kw_scopes)
    {
        // verify the keyword scope is not a dup of a declared scope
        auto iter = find(declared_scopes.begin(), declared_scopes.end(), kw_scope.first);
        if (iter == declared_scopes.end())
        {
            list.emplace_back(us::ChildScope{kw_scope.first, kw_scope.second, true, map[kw_scope.first]});
        }
    }

    for (auto cd : list)
        qDebug() << "==== cs id: " << qstr(cd.id);
    return list;
}

#define EXPORT __attribute__ ((visibility ("default")))

extern "C"
{

EXPORT
us::ScopeBase*
// cppcheck-suppress unusedFunction
UNITY_SCOPE_CREATE_FUNCTION()
{
    return new AggScope();
}

EXPORT
void
// cppcheck-suppress unusedFunction
UNITY_SCOPE_DESTROY_FUNCTION(us::ScopeBase* scope_base)
{
    delete scope_base;
}

}
