/*
 * Copyright (C) 2015 Canonical Ltd
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 3 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Kyle Nitzsche <kyle.nitzsche@canonical.com>
 *              Scott Sweeny <scott.sweeny@canonical.com>
 *
 */

#include "query.h"
#include "aggchildscope.h"
#include "i18n.h"
#include "utils.h"

#include <unity/UnityExceptions.h>

#include <QFile>
#include <QJsonParseError>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonArray>
#include <QDateTime>

using namespace unity::scopes;

QString locale_qs = QString::fromStdString(setlocale(LC_ALL, ""));
QString lang = locale_qs.left(2);
QString mid = locale_qs.left(5);
QString country = mid.right(2).toLower();

QString qstr(std::string str)
{
    return QString::fromStdString(str);
}

std::string sstr(QString str)
{
    return str.toStdString();
}

void Query::load_cardinality_settings()
{

    QString child_scopes_fn_qs = QString("%1/child_scopes.json").arg(qstr(scope_dir_));

    QFile child_scopes_f;
    child_scopes_f.setFileName(child_scopes_fn_qs);
    child_scopes_f.open(QIODevice::ReadOnly);
    QByteArray data = child_scopes_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);

    if (err.error != QJsonParseError::NoError)
    {
        qCritical() << "Failed to parse child_scopes.json data: " << err.errorString();
        return;
    }

    QJsonObject doc_o = doc.object();

    if (doc_o.contains("cardinality_settings"))
    {
        QJsonArray card_a = doc_o["cardinality_settings"].toArray();
        for (auto item : card_a)
        {
            QJsonObject item_o = item.toObject();
            qDebug() << " ===== CARD. found declared value: " << item_o["value"].toInt();
            setting_cardinalities.push_back(item_o["value"].toInt());
        }
    }
}

int Query::get_setting_cardinality()
{
    us::VariantMap settings_ = settings();
    auto it = settings_.find("cardinality");
    if (it != settings_.end())
    {
        qDebug() << "==== SETTINGS. CARD cardinality s: " << it->second.get_int();
        return setting_cardinalities.at(it->second.get_int());
    }
    return -1;
}

/*
 * Load declarations from child_scopes.json
 */
void Query::load_declarations()
{

    QString child_scopes_fn_qs = QString("%1/child_scopes.json").arg(qstr(scope_dir_));

    QFile child_scopes_f;
    child_scopes_f.setFileName(child_scopes_fn_qs);
    child_scopes_f.open(QIODevice::ReadOnly);
    QByteArray data = child_scopes_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);

    if (err.error != QJsonParseError::NoError)
    {
        qCritical() << "Failed to parse child_scopes.json data: " << err.errorString();
        return;
    }

    QJsonObject doc_o = doc.object();

    QJsonArray common_templates_a = doc_o["common_templates"].toArray();

    for (const auto item_ : common_templates_a)
    {
        QJsonObject template_o = item_.toObject();
        std::string id = template_o["id"].toString().toStdString();

        QJsonObject template__ = template_o["template"].toObject();
        QJsonDocument t_d(template__);
        std::string template_ = sstr(t_d.toJson());

        common_templates[id] = template_;
        if (template_o.contains("fallbacks"))
        {
            qDebug() << "==== FALLBACK in fallbacks, id: " << qstr(id);
            QJsonArray fallbacks_a = template_o["fallbacks"].toArray();
            std::map<std::string,std::vector<std::string>> fallbacks_m;
            std::vector<std::string> fields;
            for (const auto fb_ : fallbacks_a)
            {
                QJsonObject fb_o = fb_.toObject();
                std::string key = fb_o["key"].toString().toStdString();
                qDebug() << "==== FALLBACK key: " <<qstr(key);
                QJsonArray fields_a = fb_o["fields"].toArray();
                for (const auto f_ : fields_a)
                {
                    QJsonObject f_o = f_.toObject();
                    std::string field= f_o["field"].toString().toStdString();
                    qDebug() << "==== FALLBACK field: " <<qstr(field);
                    fields.emplace_back(field);
                }
                fallbacks_m[key] = fields;
            }
            std::shared_ptr<std::map<std::string,std::vector<std::string>>> fbs = std::make_shared<std::map<std::string,std::vector<std::string>>>(fallbacks_m);
            common_templates_fallbacks[id] = fbs;
            qDebug() << "==== load FALLB size 1: " << common_templates_fallbacks.size();
            for (auto t : common_templates_fallbacks)
            {
                qDebug() << "==== load FALLB template id : " << qstr(t.first);
                auto m_ptr = t.second;
                for (auto &v : *m_ptr)
                {
                    qDebug() << "==== load FALLB c_t_f result component " << qstr(v.first);
                    for (auto f : v.second)
                    {
                        qDebug() << "==== load  FALLB c_t_f result fallback " << qstr(f);
                    }
                }

            }
        }
    }
    QJsonObject departments_o = doc_o["departments"].toObject();
    if (doc_o.contains("departments"))
    {
        QJsonObject departments_o = doc_o["departments"].toObject();

        bool use_depts = false;
        if (departments_o.contains("do_not_use_departments"))
        {
            if (departments_o["do_not_use_departments"].toString() != "true")
            {
                using_departments = true;
                use_depts = true;
            }
            else
            {
                using_departments = false;
            }
        }
        if (use_depts)
        {
            QJsonArray declarations_a = departments_o["declarations"].toArray();
            for (const auto item_ : declarations_a)
            {
                dept d_;
                std::shared_ptr<dept> dept_ = std::make_shared<dept>(d_);
                QJsonObject dept_o = item_.toObject();
                dept_->id = dept_o["id"].toString().toStdString();
                dept_->title = _(dept_o["_title"].toString().toStdString().c_str());
                if (dept_o.contains("root"))
                {
                    if (dept_o["root"].toString() == "true")
                    {
                        dept_->root = true;
                        dept_id_of_root = dept_->id;
                    }
                }
                depts[dept_->id] = dept_;
            }
        }
    }
    QJsonArray order_a = doc_o["order"].toArray();

    for (const auto item_o_ : order_a)
    {
        QJsonObject item_o = item_o_.toObject();

        if (item_o.contains("category"))
        {
            QJsonObject cat_o = item_o["category"].toObject();
            category cat_;
            std::shared_ptr<category> cat = std::make_shared<category>(cat_);
            cat->id = cat_o["id"].toString().toStdString();

            cat->title = cat_o["_title"].toString().toStdString();
            cat->dept_id = cat_o["department"].toString().toStdString();

            categoryId_titleMsgid[cat->id] = cat->title;
            categoryIds.emplace_back(cat->id);

            if (cat_o.contains("link_to_child_specified"))
            {
                categoryId_linkToChildSpecified[cat->id] = cat_o["link_to_child_specified"].toString().toStdString();
            }
            int card = get_setting_cardinality();
            if (card > 0)
            {
                cat->cardinality = card;
            }
            // but let declaration for category override user settings
            if (cat_o.contains("cardinality"))
            {
                int cardinality_ = cat_o["cardinality"].toInt();
                if (cardinality_ > -1 )
                {
                    cat->cardinality = cardinality_;
                }
            }

            if (cat_o.contains("first_result_renderer_common_id"))
            {
                std::string id = cat_o["first_result_renderer_common_id"].toString().toStdString();
                categoryId_first_result_renderers[cat->id] = common_templates[id];
                categoryId_firstRdrs_catId2commonId[cat->id] = id;
            }
            else if (cat_o.contains("first_result_renderer"))
            {
                QJsonObject template_o = cat_o["first_result_renderer"].toObject();
                QJsonDocument t_d(template_o);
                categoryId_first_result_renderers[cat->id] = sstr(t_d.toJson());
            }
            else if (cat_o.contains("first_result_template"))
            {
                //support deprecated for backwards compatibiliity
                qWarning () << QString("%1: DEPRECATED use of 'first_result_template'. Switch to 'first_result_renderer'").arg(qstr(scope_id));
                QJsonObject template_o = cat_o["first_result_template"].toObject();
                QJsonDocument t_d(template_o);
                categoryId_first_result_renderers[cat->id] = sstr(t_d.toJson());
            }

            if (cat_o.contains("renderer_common_id"))
            {
                std::string id = cat_o["renderer_common_id"].toString().toStdString();
                cat->renderer = common_templates[id];
                categoryId_surfaceRdrs_catId2commonId[cat->id] = id;
            }
            else if (cat_o.contains("renderer"))
            {
                QJsonObject renderer_o = cat_o["renderer"].toObject();
                QJsonDocument st_d(renderer_o);
                cat->renderer = sstr(st_d.toJson());
            }

            if (cat_o.contains("search_renderer_common_id"))
            {
                std::string id = cat_o["search_renderer_common_id"].toString().toStdString();
                cat->search_renderer = common_templates[id];
            }
            else if (cat_o.contains("search_renderer"))
            {
                QJsonObject renderer_o = cat_o["search_renderer"].toObject();
                QJsonDocument st_d(renderer_o);
                cat->search_renderer = sstr(st_d.toJson());
            }

            if (cat_o.contains("scopes"))
            {
                QJsonArray scope_a = cat_o["scopes"].toArray();

                for (const auto item : scope_a)
                {
                    QJsonObject scope_o = item.toObject();
                    std::string id = scope_o["id"].toString().toStdString();
                    std::string local_id = scope_o["local_id"].toString().toStdString();

                    //TODO doc that a child_department may include LANGCODE for replacement
                    //from locale lang code, and COUNTRYCODE with locale country code
                    if (scope_o.contains("child_department"))
                    {
                        QString child_dept = scope_o["child_department"].toString();
                        if (lang.size() == 2)
                        {
                            if (child_dept.contains("LANGCODE"))
                                child_dept = child_dept.replace("LANGCODE", lang);
                                scopeLocalId_childDept_m[local_id] = child_dept.toStdString();
                        }
                        if (country.size() == 2)
                        {

                            if (child_dept.contains("COUNTRYCODE"))
                                child_dept = child_dept.replace("COUNTRYCODE", country);
                                scopeLocalId_childDept_m[local_id] = child_dept.toStdString();
                        }
                        qDebug() << qstr("=== CHDEPT. lang: %1 local %2. dept %3").arg(lang, qstr(local_id), qstr(scopeLocalId_childDept_m[local_id]));
                    }
                    if (scope_o.contains("default_query_string"))
                    {

                        std::string default_qry_str = scope_o["default_query_string"].toString().toStdString();
                        localId_defaultQryStr[local_id] = default_qry_str;
                    }

                    cat->scopes.emplace_back(std::pair<std::string,std::string>(id,local_id));
                }
            }
            if (cat_o.contains("keywords"))
            {
                QJsonArray keyword_a = cat_o["keywords"].toArray();

                for (const auto item : keyword_a)
                {
                    QJsonObject keyword_o = item.toObject();
                    std::string id = keyword_o["id"].toString().toStdString();
                    std::string local_id = keyword_o["local_id"].toString().toStdString();
                    cat->keywords.emplace_back(id);
                    cat->keyword_ids.emplace_back(std::pair<std::string,std::string>(id,local_id));
                    declared_category_keywords.emplace_back(id);
                }
            }

            declared_order.emplace_back(cat->id);
            categories.emplace_back(cat);
            id_category_m[cat->id] = cat;
        }

        if (item_o.contains("scope"))
        {
            QJsonObject scope_o = item_o["scope"].toObject();
            child_scope cs_;
            std::shared_ptr<child_scope> cs = std::make_shared<child_scope>(cs_);

            cs->id = scope_o["id"].toString().toStdString();
            cs->local_id = scope_o["local_id"].toString().toStdString();
            declared_order.emplace_back(cs->local_id);

            if (scope_o.contains("department"))
                cs->department = scope_o["department"].toString().toStdString();

            if (scope_o.contains("link_to_child"))
                if (scope_o["link_to_child"].toString() == "false")
                    cs->link_to_child = false;
            else
                cs->link_to_child = true;

            if (scope_o.contains("source_finder"))
                if (scope_o["source_finder"].toString() == "true") {
                    cs->source_finder = true;
                }

            if (scope_o.contains("category_title_use_incoming"))
            {
                cs->using_category_title_incoming = true;
            }
            else if (scope_o.contains("category_title_use_display_name"))
            {
                cs->using_category_title_display_name = true;
            }
            else
            {
                if (scope_o.contains("category_title"))
                    cs->category_title = scope_o["category_title"].toString().toStdString();

                if (scope_o.contains("_category_title"))
                    cs->category_title_ = scope_o["_category_title"].toString().toStdString();
                qDebug() << "==== TITLE_: " << qstr(cs->category_title);
                qDebug() << "==== TITLE_: " << qstr(cs->category_title_);
            }

            //TODO doc that a child_department may include LANGCODE for replacement
            //from locale lang code, and COUNTRYCODE with locale country code
            if (scope_o.contains("child_department"))
            {
                QString child_dept = scope_o["child_department"].toString();
                if (lang.size() == 2)
                {
                    if (child_dept.contains("LANGCODE"))
                        child_dept = child_dept.replace("LANGCODE", lang);
                }
                if (country.size() == 2)
                {

                    if (child_dept.contains("COUNTRYCODE"))
                        child_dept = child_dept.replace("COUNTRYCODE", country);
                }
                cs->child_department = child_dept.toStdString();
            }

            if (scope_o.contains("child_category"))
                cs->child_category = scope_o["child_category"].toString().toStdString();

            if (scope_o.contains("child_category_max_results"))
            {
                cs->child_category_max_results = (int)scope_o["child_category_max_results"].toInt();
                cs->using_child_category_max_results = true;
            }
            else if (scope_o.contains("child_category"))
                cs->using_child_category_max_results = true;

            int card = get_setting_cardinality();
            if (card > 0)
            {
                cs->cardinality = card;
            }
            // but let declaration for category override user settings
            if (scope_o.contains("cardinality"))
            {
                int cardinality_ = scope_o["cardinality"].toInt();
                if (cardinality_ > -1)
                {
                    cs->cardinality = cardinality_;
                }
            }

            if (scope_o.contains("first_result_renderer_common_id"))
            {
                std::string id = scope_o["first_result_renderer_common_id"].toString().toStdString();
                cs->first_result_template = common_templates[id];
            }
            else if (scope_o.contains("first_result_renderer"))
            {
                QJsonObject template_o = scope_o["first_result_renderer"].toObject();
                QJsonDocument t_d(template_o);
                cs->first_result_template = sstr(t_d.toJson());
            }
            else if (scope_o.contains("first_result_template"))
            {
                //support deprecated for backwards compatibiliity
                qWarning () << QString("%1: DEPRECATED use of 'first_result_template'. Switch to 'first_result_renderer'").arg(qstr(scope_id));
                QJsonObject template_o = scope_o["first_result_template"].toObject();
                QJsonDocument t_d(template_o);
                cs->first_result_template = sstr(t_d.toJson());
            }

            if (scope_o.contains("renderer_common_id"))
            {
                std::string id = scope_o["renderer_common_id"].toString().toStdString();
                cs->surface_template = common_templates[id];
                cs->overriding_surface_template = true;
            }
            else if (scope_o.contains("renderer"))
            {
                QJsonObject surface_template_o = scope_o["renderer"].toObject();
                QJsonDocument st_d(surface_template_o);
                cs->surface_template = sstr(st_d.toJson());
                cs->overriding_surface_template = true;
            }
            else if (scope_o.contains("surface_template"))
            {
                //support deprecated for backwards compatibility
                qWarning () << QString("%1: DEPRECATED use of 'surface_template'. Switch to 'surface_renderer'").arg(qstr(scope_id));
                QJsonObject surface_template_o = scope_o["surface_template"].toObject();
                QJsonDocument st_d(surface_template_o);
                cs->surface_template = sstr(st_d.toJson());
                cs->overriding_surface_template = true;
            }

            if (scope_o.contains("search_renderer_common_id"))
            {
                qDebug() << "==== SEARCH red com fouund";
                std::string id = scope_o["search_renderer_common_id"].toString().toStdString();
                cs->search_template = common_templates[id];
                cs->overriding_search_template = true;
            }
            else if (scope_o.contains("search_renderer"))
            {
                QJsonObject search_template_o = scope_o["search_renderer"].toObject();
                QJsonDocument search_d(search_template_o);
                cs->search_template = sstr(search_d.toJson());
                cs->overriding_search_template = true;
            }
            else if (scope_o.contains("search_template"))
            {
                //support deprecated for backwards compatibility
                qWarning () << QString("%1: DEPRECATED use of 'search_template'. Switch to 'search_renderer'").arg(qstr(scope_id));
                QJsonObject search_template_o = scope_o["search_template"].toObject();
                QJsonDocument search_d(search_template_o);
                cs->search_template = sstr(search_d.toJson());
                cs->overriding_search_template = true;
            }

            if (scope_o.contains("swap_result_attributes"))
            {
                QJsonArray swaps_a = scope_o["swap_result_attributes"].toArray();
                for (const auto swap_o_ : swaps_a)
                {
                    QJsonObject swap_o = swap_o_.toObject();
                    std::pair<std::string,std::string> swap;
                    swap.first = swap_o["incoming_result_key"].toString().toStdString();
                    swap.second = swap_o["new_result_key"].toString().toStdString();
                    std::shared_ptr<std::pair<std::string,std::string>> swap_ = std::make_shared<std::pair<std::string,std::string>>(swap);
                    cs->swap_result_attributes.emplace_back(swap_);
                }

            }

            if (scope_o.contains("result_category_id_to_common_template"))
            {
                QJsonArray items_a = scope_o["result_category_id_to_common_template"].toArray();
                for (const auto item_o_ : items_a)
                {
                    QJsonObject item_o = item_o_.toObject();
                    std::string res_cat_id = item_o["result_category_id"].toString().toStdString();
                    std::string common_template = item_o["common_template"].toString().toStdString();
                    cs->result_category_id_to_common_template[res_cat_id] = common_template;
                }
            }

            if (scope_o.contains("default_query_string"))
            {
                QString default_query_string = scope_o["default_query_string"].toString();
                localId_defaultQryStr[cs->local_id] = default_query_string.toStdString();
            }

            if (scope_o.contains("only_in_search") && scope_o["only_in_search"] == "true")
            {
                cs->only_in_search = true;
                cs->overriding_search_template = true;
            } else {
                cs->only_in_search = false;
            }
            
            child_scopes.emplace_back(cs);
            child_scopes_m[cs->local_id] = cs;
            type_ids_m[cs->local_id].emplace_back(cs->local_id);
        }
        else if (item_o.contains("keyword"))
        {
            keyword keyword_;
            QJsonObject keyword_o = item_o["keyword"].toObject();
            std::string keyword_id = keyword_o["keyword"].toString().toStdString();
            keyword_.id = keyword_id;

            declared_keywords.emplace_back(keyword_id);
            declared_order.emplace_back(keyword_id);

            int card = get_setting_cardinality();
            if (card > 0)
            {
                keyword_.cardinality = card;
            }
            // but let declaration for category override user settings
            if (keyword_o.contains("cardinality"))
            {
                int keyword_cardinality_ = keyword_o["cardinality"].toInt();
                if (keyword_cardinality_ > -1)
                {
                    keyword_.cardinality = keyword_cardinality_;
                }
            }

            if (using_departments)
            {
                std::string dept_title_msgid = keyword_o["_department_title"].toString().toStdString();
                keyword_deptTitleMsgid[keyword_id] = dept_title_msgid;
                if (keyword_o.contains("department"))
                {

                    std::string kwdept = keyword_o["department"].toString().toStdString();
                    keyword_deptId[keyword_id] = kwdept;
                    keyword_.dept_id = kwdept;
                }
                else
                    if (dept_id_of_root_keywords_only.empty())
                        dept_id_of_root_keywords_only = keyword_id;
            }
            std::string category;
            if (keyword_o.contains("shared_category"))
            {
                //TODO, remove support for shared_category value other than "true"
                if (keyword_o["shared_category"] == "true")
                    category = keyword_id;
                else
                    category = keyword_o["shared_category"].toString().toStdString();
                std::string title_msgid = keyword_o["_shared_category_title"].toString().toStdString();
                shared_keyword_cat_titlemsgid[category] = title_msgid;
                shared_keyword_cats[keyword_id] = category;
                shared_keywords.emplace_back(keyword_id);
            }
            if (keyword_o.contains("first_result_renderer_common_id"))
            {
                std::string id = keyword_o["first_result_renderer_common_id"].toString().toStdString();
                keyword_catname_first_result_renderers[category] = common_templates[id];
            }
            else if (keyword_o.contains("first_result_renderer"))
            {
                QJsonObject template_o = keyword_o["first_result_renderer"].toObject();
                QJsonDocument t_d(template_o);
                keyword_catname_first_result_renderers[category] = sstr(t_d.toJson());
            }
            else if (keyword_o.contains("first_result_template"))
            {
                //support deprecated for backwards compatibiliity
                qWarning () << QString("%1: DEPRECATED use of 'first_result_template'. Switch to 'first_result_renderer'").arg(qstr(scope_id));
                QJsonObject template_o = keyword_o["first_result_template"].toObject();
                QJsonDocument t_d(template_o);
                keyword_catname_first_result_renderers[category] = sstr(t_d.toJson());
            }
            if (keyword_o.contains("category_renderer_common_id"))
            {
                std::string id = keyword_o["category_renderer_common_id"].toString().toStdString();
                keyword_catname_renderer[category] = common_templates[id];
            }
            else if (keyword_o.contains("category_renderer"))
            {
                QJsonObject renderer_o = keyword_o["category_renderer"].toObject();
                QJsonDocument rdr_d(renderer_o);
                keyword_catname_renderer[category] = sstr(rdr_d.toJson());
            }
            if (keyword_o.contains("search_renderer_common_id"))
            {
                std::string id = keyword_o["search_renderer_common_id"].toString().toStdString();
                keyword_catname_search_renderer[category] = common_templates[id];
            }
            else if (keyword_o.contains("search_renderer"))
            {
                QJsonObject renderer_o = keyword_o["search_renderer"].toObject();
                QJsonDocument rdr_d(renderer_o);
                keyword_catname_search_renderer[category] = sstr(rdr_d.toJson());
            }
            if (keyword_o.contains("link_to_child"))
            {
                if (keyword_o["link_to_child"].toString() == "true")
                    keyword_.link_to_child = true;
            }
            if (keyword_o.contains("exclude_scopes"))
            {
                  QJsonArray scope_a = keyword_o["exclude_scopes"].toArray();
  
                  for (const auto item : scope_a)
                  {
                      QJsonObject scope_o = item.toObject();
                      std::string id = scope_o["id"].toString().toStdString();
                      keyword_.exclude_scopes.emplace_back(id);
                  }
            }
            
            std::shared_ptr<keyword> kw_ptr = std::make_shared<keyword>(keyword_);
            keywords.emplace_back(kw_ptr);
            id_keyword_map[keyword_id]  = kw_ptr;
        }
    }
    //for (auto str : declared_order)
    //    qDebug() << "==== ORDER. declared scope: " << qstr(str);
}

void Query::set_scope_order()
{
    for (std::string id : declared_order)
    {
    qDebug() << "==== SET ORDER. DECLARED order id: " << qstr(id);
        for (auto pr : type_ids_m)
        {
            if (pr.first == id)
            {
                for (std::string id_ : pr.second)
                {
                    scopes_ordered.emplace_back(id_);
                    qDebug() << "==== SET ORDER DECLARED adding to scopes_ordered id: " << qstr(id_);
                }
            }
        }
    }
}

/*
 * makes the AggChildScope instances using values and options
 * declared in json, else default values from child_scope struct
 */

void Query::make_declared_scopes()
{
    //iterate declared child scopes
    for (std::shared_ptr<child_scope> child : child_scopes)
    {
        AggChildScope ch(child->id);

        ch.set_local_id(child->local_id);

        localId_id_m[child->local_id] = child->id;

        if (child->using_category_title_incoming)
        {
            ch.set_category_title_incoming(true);
        }
        else  if (child->using_category_title_display_name)
        {
            ch.set_category_title_display_name(true);
        }
        else
        {
            if (child->category_title_ != "")
            {
               ch.set_category_title(_(child->category_title_.c_str()));
            }
            else
               ch.set_category_title(child->category_title);
        }

        if (child->child_department != "")
           ch.set_child_department(child->child_department);

        if (child->department != "")
           ch.set_department(child->department);

        ch.set_category_link_to_child(child->link_to_child);

        if (child->child_category != "")
           ch.set_child_category(child->child_category);

        if (child->using_child_category_max_results)
            ch.set_child_category_max_results(child->child_category_max_results);

        int card = get_setting_cardinality();
        if (card > 0)
            ch.set_cardinality(card);
        if (child->cardinality >= -1) // let declared override user settings
            ch.set_cardinality(child->cardinality);

        if (!child->first_result_template.empty())
            ch.set_first_result_template(child->first_result_template);

        if (child->overriding_surface_template)
        {
            ch.set_surface_template(child->surface_template);
        }
        else
            ch.set_override_surface_template(false);

        if (child->overriding_search_template)
        {
            ch.set_search_template(child->search_template);
        }
        else
            ch.set_override_search_template(false);

        ch.set_only_in_search(child->only_in_search);

        ch.set_source_finder(child->source_finder);
        qWarning () << QString("%1: ADDING DECLARED child scope: %2, local_id: %3").arg(qstr(scope_id), qstr(ch.id()), qstr(ch.local_id()));
        scopes_m[ch.local_id()] = std::make_shared<AggChildScope>(ch);
        current_scopes.emplace_back(ch.local_id());

    }
    return;
}

void Query::create_departments(us::SearchReplyProxy const &reply_)
{
    //create depts for declared keywords
    for (auto item : keyword_deptTitleMsgid)
    {
        auto iter = keyword_deptId.find(item.first);
        if (iter != keyword_deptId.end())
        {
            continue;
        }
        dept d;
        d.title = _(item.second.c_str());
        d.id = item.first;
        depts[d.id] = std::make_shared<dept>(d);
    }

    us::DepartmentList depts_;
    us::CannedQuery cannedQuery("aggregator");
    if (dept_id_of_root.empty())//means no declared scopes, only keywords
        dept_id_of_root = dept_id_of_root_keywords_only;
    cannedQuery.set_department_id(dept_id_of_root);
    us::Department::SPtr topDept;
    topDept = move(us::Department::create("", cannedQuery, depts[dept_id_of_root]->title));
    for ( auto & item : depts)
    {
        if (item.first == dept_id_of_root)
            continue;
        us::CannedQuery cannedQuery(item.first);
        cannedQuery.set_department_id(item.first);
        qDebug()<<"=== SUB DEPT adding: " << qstr(item.first);
        us::Department::SPtr aDept = move(us::Department::create(item.first, cannedQuery, item.second->title));
        depts_.insert(depts_.end(), aDept);
    }

    topDept->set_subdepartments(depts_);
    reply_->register_departments(topDept);
}

void Query::make_keyword_scopes()
{
    for (auto const& keyword_child : current_child_scopes)
    {

        if (!keyword_child.enabled)
            continue;

        std::vector<std::string>::iterator iter;
        iter = find(current_scopes.begin(), current_scopes.end(), localId_id_m[keyword_child.id]);
        if (iter != current_scopes.end()) // don't add as keyword if already added as declared
            continue;

        AggChildScope keych(keyword_child.id);
        QString time_str = QDateTime::currentDateTimeUtc().toString();
        keych.set_local_id(keyword_child.id + time_str.toStdString());
        keych.set_keyword_scope(true);
        dept d;
        d.id = keyword_child.id;
        std::string id = keyword_child.id;
        bool found = false;
        for (auto kw : declared_keywords)
        {
            qDebug() << "==== %1 KEYWORD: " <<  qstr(kw);
            auto found_scopes = agg_scope_->registry()->list_if([kw, id](us::ScopeMetadata const& md)
            {
                auto keywords = md.keywords();
                if (md.scope_id() == id)
                    if (keywords.find(kw) != keywords.end())
                        return true;
                    else
                        return false;
                return false;
            });
            if (found_scopes.size() == 0 )
                continue;
            type_ids_m[kw].emplace_back(id);
            d.keyword = kw;
            keych.set_department(kw);
            keych.set_keyword(kw);
            keych.set_category_link_to_child(id_keyword_map[kw]->link_to_child);
            qWarning() << "==== ID KW 1: " << qstr(keych.keyword());

            bool a_dup = false;
            // if this keyword scope is already a declared scope in the current dept, don't add it
            if (using_departments)
            {
                std::string curr_dept_id = query().department_id();

                //make sure we load correct dept id if we specify department property in keyword object
                if (!curr_dept_id.empty() && keyword_deptId[kw] != curr_dept_id) {
                    continue;
                }

                for (std::shared_ptr<child_scope> cs : child_scopes)
                {
                    if (keyword_child.id != cs->id)
                        continue;
                    if (curr_dept_id == "")
                    {
                        if (cs->department == dept_id_of_root)
                        {
                            qDebug()<< "====  not ADDING to cur dept, IS a DUP:" << qstr(keyword_child.id);
                            a_dup =true;
                            break;
                        }
                    }
                    else if (cs->department == curr_dept_id)
                    {
                        qDebug()<< "====  not ADDING to cur dept, IS a DUP:" << qstr(keyword_child.id);
                        a_dup =true;
                        break;
                    }
                }
                if (a_dup)
                    continue;
            }
            if (depts.find(kw) == depts.end())
            {
                if (keyword_deptId.find(kw) != keyword_deptId.end())
                {
                    // keyword is declared to use declared department, so use it
                    keych.set_department(keyword_deptId[kw]);
                }
                else
                {
                    depts[kw] = std::make_shared<dept>(d);
                }
            }
            int idx = -1;
            auto iter = find(shared_keywords.begin(), shared_keywords.end(), kw);
            if (iter != shared_keywords.end())
            {
                qWarning() << "==== ID JUST set kw: "<<  qstr(keych.keyword());
                qDebug() << QString("==== set ky scope kw %1 as a a shared cat in kw: %2").arg(qstr(keych.id()), qstr(kw));
                keych.set_keyword_scope_shared_cat_name(shared_keyword_cats[kw]);
                //initially the category is set to false meaning there have been
                //no results yet for it and it has not yet been registered
                sharedcat_catregistered[shared_keyword_cats[kw]] = false;
                // set more default values
                keywordSharedCat_isFirstResult[kw] = true;
                qDebug() << "===== UTILs. kw: " << qstr(kw);
                qDebug() << "===== UTILs. keywordSharedCat_isFirstResult[kw] " << keywordSharedCat_isFirstResult[kw];
                keywordSharedCat_isSecondResult[kw] = false;
                keych.set_surface_template(keyword_catname_renderer[kw]);
            }
            found = true;
            //this break ensures the scope is associated with *first* decared keyword, thus
            //ensuring keyword scopes follow the overall declared order
            break;
        }
        if (!found)
            continue;
        std::string curr_dept_id = query().department_id();
        if (curr_dept_id.empty())
            curr_dept_id = "root";
        qWarning() << "==== ID 2 KEYWORD: " << qstr(keych.keyword());
        qWarning () << QString("%1: ADDING KEYWORD child scope: %2 by keyword %3 to department: %4").arg(qstr(scope_id), qstr(keych.id()), qstr(keych.keyword()), qstr(curr_dept_id));
        scopes_m[keych.id()] = std::make_shared<AggChildScope>(keych);
        localId_id_m[keych.id()] = keych.id();
        current_scopes.emplace_back(keych.id());
    }
}

void Query::make_categories()
{
    for (auto cat : categories)
    {
        qDebug() << "==== MKCAT id: " << qstr(cat->id);
        qDebug() << "==== MKCAT dept id: " << qstr(cat->dept_id);
        categoryId_isFirstResult[cat->id] = true;
        categoryId_isSecondResult[cat->id] = false;

        for (std::pair<std::string,std::string> ids : cat->scopes) //id, local_id
        {
            qDebug() << "==== MKCAT SCOPE scope id"<< qstr(ids.first);
            qDebug() << "==== MKCAT SCOPE scope loca_id"<< qstr(ids.second);
            AggChildScope scope(ids.first);
            scope.set_category_scope(true);
            scope.set_category_id(cat->id);
            scope.set_local_id(ids.second);
            if (!scopeLocalId_childDept_m[ids.second].empty())
            {
                //qDebug() << "==== CHDEPT id: " << qstr(scope->id);
                //qDebug() << "==== CHDEPT: " << qstr(scope->child_dept);
                scope.set_child_department(scopeLocalId_childDept_m[ids.second]);
            }
            scope.set_surface_template(cat->renderer);
            scope.set_search_template(cat->search_renderer);
            if (using_departments)
                scope.set_department(cat->dept_id);
            qDebug() << "==== MKCAT SCOPE cat dept_id"<< qstr(cat->dept_id);
            qDebug() << "==== MKCAT SCOPE scope dept_id"<< qstr(scope.department());
/*
            if (categoryId_firstRdrs_catId2commonId.find(cat->id) != categoryId_firstRdrs_catId2commonId.end())
            {

                scope.set_first_common_template_id(categoryId_firstRdrs_catId2commonId[cat->id]);
            }
            if (categoryId_surfaceRdrs_catId2commonId.find(cat->id) != categoryId_surfaceRdrs_catId2commonId.end())
            {

                scope.set_surface_common_template_id(categoryId_surfaceRdrs_catId2commonId[cat->id]);
            }
*/
            scope.set_keyword_scope(false);
            scopes_m[scope.local_id()] = std::make_shared<AggChildScope>(scope);
            type_ids_m[cat->id].emplace_back(scope.local_id());
            localId_id_m[scope.local_id()] = scope.id();
            current_scopes.emplace_back(scope.local_id());
            qWarning () << QString("%1: ADDING CATEGORY SCOPE child scope: %2 to dept %3").arg(qstr(scope_id), qstr(scope.local_id()), qstr(scope.department()));
        }


        for (auto const& keyword_child : current_child_scopes)
        {
            qDebug() << "==== CATKW id START ";
            if (!keyword_child.enabled)
                continue;

            std::string id = keyword_child.id;
            AggChildScope keych(id);
            for (auto kw : cat->keywords)
            {
                qDebug() << "==== CATKW child id: " <<  qstr(id);
                qDebug() << "==== CATKW id: " <<  qstr(cat->id);
                qDebug() << "==== CATKW KW: " <<  qstr(kw);
                auto found_scopes = agg_scope_->registry()->list_if([kw,id](us::ScopeMetadata const& md)
                {
                    auto keywords = md.keywords();
                    if (md.scope_id() == id && keywords.find(kw) != keywords.end())
                        return true;
                    else
                        return false;
                });
                if (found_scopes.size() == 0 )
                    continue;

                qDebug() << qstr("==== CATKW ID found: %1 %2").arg(qstr(kw),qstr(id));
                AggChildScope keych(id);
                std::string local_id = id + ":category=" + cat->id + ":keyword=" +  kw;
                keych.set_local_id(local_id);
                keych.set_category_link_to_child(false);//TODO revisit
                keych.set_category_scope(true);
                keych.set_category_id(cat->id);
                qDebug() << "==== CATKW adding id: " << qstr(id);
                qDebug() << "==== CATKW adding kw: " << qstr(kw);
                type_ids_m[cat->id].emplace_back(keych.local_id());
                keych.set_keyword(kw);
                qDebug() << "==== CATKW: " << qstr(keych.keyword());
                keych.set_department(cat->dept_id);
                keych.set_department(cat->dept_id);
                keych.set_cardinality(cat->cardinality);

                // set more default values
                categoryId_isFirstResult[kw] = true;
                categoryId_isSecondResult[kw] = false;
                keych.set_surface_template(cat->renderer);


                if (categoryId_firstRdrs_catId2commonId.find(cat->id) != categoryId_firstRdrs_catId2commonId.end())
                {

                    keych.set_first_common_template_id(categoryId_firstRdrs_catId2commonId[cat->id]);
                }
                if (categoryId_surfaceRdrs_catId2commonId.find(cat->id) != categoryId_surfaceRdrs_catId2commonId.end())
                {

                    keych.set_surface_common_template_id(categoryId_surfaceRdrs_catId2commonId[cat->id]);
                }

                qWarning () << QString("%1: ADDING CATEGORY KEYWORD child scope: %2, by keyword %3, to department: %4").arg(qstr(scope_id), qstr(keych.id()), qstr(kw), qstr(cat->dept_id));
                scopes_m[keych.local_id()] = std::make_shared<AggChildScope>(keych);
                localId_id_m[keych.local_id()] = keych.id();
                current_scopes.emplace_back(keych.local_id());
                current_scopeIds.emplace_back(id);
            }
        }
    }
}

void Query::handle_current_child_scopes(bool empty_search, us::SearchReplyProxy const& upstream_reply)
{
    // Cycle through child scopes (in current_scopes) and, if enabled in settings, complete set up
    try
    {
        for (const std::string & local_id : current_scopes)
        {
            qDebug() << "==== CURRENT. local_id: " << qstr(local_id);
            std::string id = localId_id_m[local_id];
            qDebug() << "==== CURRENT. id: " << qstr(id);
            std::shared_ptr<AggChildScope> scope = scopes_m[local_id];

            scope->set_proxy(registry_);

            if (!scope->exists()) {
                qWarning () << QString("%1: scope is NOT REGISTERED, skipping: %2").arg(qstr(scope_id), qstr(id));
                auto myID = QString::fromStdString(id);
                myID = myID.left(myID.indexOf("_"));
                if (query_store_ == "")
                {
                    query_store_.append("name:");
                }
                else
                {
                    query_store_.append("+");
                }
                query_store_.append(myID.toStdString());
            }
            else
            {
                if (!scope->source_finder())
                    has_one_source = true;
            }
        }
    }
    catch (unity::Exception const& e)
    {
        qWarning() << "NotFoundException: " << e.what();
    }
}

void Query::noSources(unity::scopes::SearchReplyProxy const& reply) {
    std::string MESS_GRID = R"(
        {
            "schema-version" : 1,
            "template" : {
                "category-layout" : "grid",
                "card-layout": "horizontal",
                "card-size": "small",
                "non-interactive": true
            },
            "components" : {
                "title" : "title"
            }
        }
    )";

    CategoryRenderer emptyCat(MESS_GRID);
    auto messGrid = reply->register_category("emptyMess", _("Welcome"), "", emptyCat);
    CategorisedResult res(messGrid);
    res.set_uri("http://ubuntu.com");
    res["title"]=_("To start using this scope please install sources from the list below");
    reply->push(res);
}

void Query::load_hints()
{
    // try to open hints.json and get vals, else no hints
    QString hints_fn_qs = QString("%1/hints.json").arg(qstr(scope_dir_));

    QFile hints_f;
    hints_f.setFileName(hints_fn_qs);
    if (!hints_f.exists())
        return;
    hints_f.open(QIODevice::ReadOnly);
    QByteArray data = hints_f.readAll();

    QJsonParseError err;
    QJsonDocument doc = QJsonDocument::fromJson(data, &err);

    if (err.error != QJsonParseError::NoError)
    {
        qWarning() << "Failed to parse hints.json data: " << err.errorString();
        return;
    }

    QJsonObject doc_o = doc.object();

    if (doc_o.contains("message_on_no_network"))
    {
        if (doc_o["message_on_no_network"].toString().toStdString() == "true")
            no_network_msg = true;
    }
    if (doc_o.contains("display_hints_on_no_network"))
    {
        if (doc_o["display_hints_on_no_network"].toString().toStdString() == "true")
            no_network_msg_hints = true;
    }

    if (doc_o.contains("message_on_no_location"))
    {
        if (doc_o["message_on_no_location"].toString().toStdString() == "true")
            no_location_msg = true;
    }
    if (doc_o.contains("display_hints_on_no_location"))
    {
        if (doc_o["display_hints_on_no_location"].toString().toStdString() == "true")
            no_location_msg_hints = true;
    }

    if (doc_o.contains("using_hints"))
    {
        if (doc_o["using_hints"].toString().toStdString() == "true")
        {
            uses_hints = true;
        }
    }

    if (doc_o.contains("display_hints_on_firstuse"))
        if (doc_o["display_hints_on_firstuse"].toString().toStdString() == "true")
            hints_quickstart = true;

    if (doc_o.contains("hints_scope_id"))
        HINTS_SCOPE_ID = doc_o["hints_scope_id"].toString().toStdString();
    else
        uses_hints = false;

    if (doc_o.contains("hints_this_scope"))
        HINTS_THIS_SCOPE = doc_o["hints_this_scope"].toString().toStdString();
    else
        uses_hints = false;

    if (doc_o.contains("hints_hide"))
        HINTS_HIDE = doc_o["hints_hide"].toString().toStdString();
    else
        uses_hints = false;

    return;
}


bool Query::hints_exists()
{
    auto scopes = registry_->list();
    us::MetadataMap::const_iterator scope;
    scope = scopes.find(HINTS_SCOPE_ID);
    if (scope != scopes.end()) {
        auto meta = registry_->get_metadata(HINTS_SCOPE_ID);
        hints_scope = meta.proxy();
        return true;
    }
    return false;
}

void Query::display_hints_quickstart(us::SearchReplyProxy const& upstream_reply_)
{
    Category::SCPtr hint_cat;
    SearchMetadata metadata(search_metadata());

    auto reply = std::make_shared<ResultForwarder>(upstream_reply_,
        [this, &hint_cat](CategorisedResult& res) -> bool {
                                return true;
    });

    subsearch(hints_scope, "", HINTS_THIS_SCOPE, FilterState(),
              metadata, reply);
}

bool Query::is_hints_quickstart_dismissed(){

    auto filepath1 = QString::fromStdString(cache_dir_) +"/" + hints_is_hidden;
    auto filepath2 = QString::fromStdString(cache_dir_) +"/" + firstboot;
    if (QFile::exists(filepath1) || QFile::exists(filepath2))
        return true;
    else
        return false; //don't show hints as fall back
}

void Query::dismiss_hints_quickstart() {
    auto filepath= QString::fromStdString(cache_dir_) +"/" + hints_is_hidden;
    QFile file(filepath);
    file.open(QIODevice::WriteOnly);
    QString msg = QString("Timestamp for %1").arg(qstr(HINTS_THIS_SCOPE));
    file.write(sstr(msg).c_str(), sstr(msg).size());
    file.close();
}
